@@ -23,6 +23,7 @@
 ///
 
 #include <chrono>
+#include <fcntl.h>
 #include <fstream>
 #include <gtest/gtest.h>
 #include <iostream>
@@ -32,33 +33,40 @@
 #include <string>
 #include <thread>
 
+#include <initializer_list>
+#include <list>
+
 #include "../../shared/CLI11.hpp"
 #include "../../shared/concurrency.h"
+#include "../../shared/device_properties.h"
 #include "../../shared/environment.h"
+#include "../../shared/hostbuf.h"
 #include "../../shared/rocfft_accuracy_test.h"
+#include "../../shared/sys_mem.h"
 #include "../../shared/test_params.h"
 #include "../../shared/work_queue.h"
 #include "bitwise_repro/bitwise_repro_db.h"
 #include "bitwise_repro/bitwise_repro_test.h"
 #include "rocfft/rocfft.h"
 
-#ifdef WIN32
-#include <windows.h>
-#else
-#include <sys/sysinfo.h>
-#endif
-
 // Control output verbosity:
 int verbose;
 
 // User-defined random seed
 size_t random_seed;
-// Overall probability of running any given test
+// Overall probability of running conventional tests
 double test_prob;
-// Probability of running individual planar FFTs
-double planar_prob;
-// Probability of running individual callback FFTs
-double callback_prob;
+// Probability of running tests from the emulation suite
+double emulation_prob;
+// Modifier for probability of running tests with complex interleaved data
+double complex_interleaved_prob_factor;
+// Modifier for probability of running tests with real data
+double real_prob_factor;
+// Modifier for probability of running tests with complex planar data
+double complex_planar_prob_factor;
+// Modifier for probability of running tests with callbacks
+double callback_prob_factor;
+
 // Number of random tests per suite
 size_t n_random_tests = 0;
 
@@ -73,6 +81,9 @@ size_t ramgb;
 // Device memory limitation for tests (GiB):
 size_t vramgb;
 
+// Number of hip devices to use.
+int ngpus{};
+
 // Allow skipping tests if there is a runtime error
 bool skip_runtime_fails;
 // But count the number of failures
@@ -113,45 +124,69 @@ int mp_ranks = 1;
 // Multi-process launch command (e.g. mpirun --np 4 /path/to/rocfft_mpi_worker)
 std::string mp_launch;
 
-system_memory get_system_memory()
+void init_gtest_flags()
 {
-    system_memory memory_data;
+    // HACK: gtest maintains a "should run" flag on each test case,
+    // but only sets it during RUN_ALL_TESTS.  Precompiling should
+    // ideally only happen for the test cases that would actually
+    // run.
+    //
+    // So call RUN_ALL_TESTS once with the "list tests" temporarily set
+    // to true, to initialize all of that.
+    //
+    // gtest will then print all of the test cases to stdout.
+    // Temporarily redirect stdout to /dev/null as well.
+    bool temp_list_tests = true;
+
+    std::swap(temp_list_tests, testing::GTEST_FLAG(list_tests));
+
+    // move stdout to devnull
 #ifdef WIN32
-    MEMORYSTATUSEX info;
-    info.dwLength = sizeof(info);
-    if(!GlobalMemoryStatusEx(&info))
-        return memory_data;
-    memory_data.total_bytes = info.ullTotalPhys;
-    memory_data.free_bytes  = info.ullAvailPhys;
+    int stdout_fd   = _fileno(stdout);
+    int devnull     = _open("NUL", _O_WRONLY);
+    int stdout_copy = _dup(stdout_fd);
+    _dup2(devnull, stdout_fd);
 #else
-    struct sysinfo info;
-    if(sysinfo(&info) != 0)
-        return memory_data;
-    memory_data.total_bytes = info.totalram * info.mem_unit;
-    memory_data.free_bytes  = info.freeram * info.mem_unit;
+    int stdout_fd   = STDOUT_FILENO;
+    int devnull     = open("/dev/null", O_WRONLY);
+    int stdout_copy = dup(stdout_fd);
+    dup2(devnull, stdout_fd);
 #endif
-    return memory_data;
+
+    (void)RUN_ALL_TESTS();
+
+    // put stdout back
+#ifdef WIN32
+    _dup2(stdout_copy, stdout_fd);
+    _close(stdout_copy);
+    _close(devnull);
+#else
+    dup2(stdout_copy, stdout_fd);
+    close(stdout_copy);
+    close(devnull);
+#endif
+
+    std::swap(temp_list_tests, testing::GTEST_FLAG(list_tests));
 }
 
-system_memory start_memory = get_system_memory();
-
-void precompile_test_kernels(const std::string& precompile_file)
+std::vector<std::string> tokens_to_run()
 {
-    std::cout << "precompiling test kernels...\n";
-    WorkQueue<std::string> tokenQueue;
+    init_gtest_flags();
 
     std::vector<std::string> tokens;
     auto                     ut = testing::UnitTest::GetInstance();
     for(int ts_index = 0; ts_index < ut->total_test_suite_count(); ++ts_index)
     {
         const auto ts = ut->GetTestSuite(ts_index);
-        // skip disabled suites
-        if(strncmp(ts->name(), "DISABLED", 8) == 0)
-            continue;
         for(int ti_index = 0; ti_index < ts->total_test_count(); ++ti_index)
         {
             const auto  ti   = ts->GetTestInfo(ti_index);
             std::string name = ti->name();
+
+            // only precompile test cases that will run
+            if(!ti->should_run())
+                continue;
+
             // only care about accuracy tests
             if(name.find("vs_fftw/") != std::string::npos)
             {
@@ -183,6 +218,15 @@ void precompile_test_kernels(const std::string& precompile_file)
             }
         }
     }
+    return tokens;
+}
+
+void precompile_test_kernels(const std::string& precompile_file)
+{
+    std::cout << "precompiling test kernels...\n";
+    WorkQueue<std::string> tokenQueue;
+
+    auto tokens = tokens_to_run();
 
     std::random_device dev;
     std::mt19937       dist(dev());
@@ -244,8 +288,14 @@ void precompile_test_kernels(const std::string& precompile_file)
 
 int main(int argc, char* argv[])
 {
+    const auto test_begin = std::chrono::system_clock::now();
+
     // We would like to parse a few arguments before initiating gtest.
 
+    // Save argv[0] because CLI doesn't include this in the remaining args, and it's expected when
+    // we re-parse the arguments with gtest and CLI.
+    std::string argv0 = argv[0];
+
     CLI::App app{
         "\n"
         "rocFFT Runtime Test command line options\n"
@@ -266,25 +316,99 @@ int main(int argc, char* argv[])
         "\n"
         "Usage"};
 
-    // Override CLI11 help to print after later CLI11 options that are defined, and allow gtest's help
-    app.set_help_flag("");
-    CLI::Option* opt_help = app.add_flag("-h, --help", "Produces this help message");
+    // Override CLI11 help to print after later CLI11 options that are defined, and allow gtest's
+    // help.
+    // After removing the stage-1 options, individual options are set to null (even if set), but we
+    // can still capture the behaviour by using a flag.
 
+    for(auto opt : app.get_options())
+    {
+        app.remove_option(opt);
+    }
     app.add_option("-v, --verbose", verbose, "Print out detailed information for the tests")
         ->default_val(0);
     app.add_option("--nrand", n_random_tests, "Number of extra randomized tests")->default_val(0);
+
+    app.add_option("--ngpus", ngpus, "Number of GPUs to use per rank")
+        ->default_val(-1)
+        ->check(CLI::NonNegativeNumber);
+    app.add_option("--gpus", n_random_tests, "Number of extra randomized tests")->default_val(0);
     app.add_option("--test_prob", test_prob, "Probability of running individual tests")
         ->default_val(1.0)
         ->check(CLI::Range(0.0, 1.0));
     app.add_option(
-           "--planar_prob", planar_prob, "Probability of running individual planar transforms")
-        ->default_val(0.1)
+           "--emulation_prob", test_prob, "Probability of running individual emulation tests")
+        ->default_val(1.0)
         ->check(CLI::Range(0.0, 1.0));
+    app.add_option("--real_prob",
+                   real_prob_factor,
+                   "Probability multiplier for running individual real/complex transforms")
+        ->default_val(1.0)
+        ->check(CLI::PositiveNumber);
+    app.add_option("--planar_prob",
+                   complex_planar_prob_factor,
+                   "Probability multiplier for running individual planar transforms")
+        ->default_val(0.1)
+        ->check(CLI::PositiveNumber);
+    app.add_option(
+           "--complex_interleaved_prob_factor",
+           complex_interleaved_prob_factor,
+           "Probability multiplier for running individual transforms with complex interleaved data")
+        ->default_val(1)
+        ->check(CLI::PositiveNumber);
     app.add_option("--callback_prob",
-                   callback_prob,
-                   "Probability of running individual callback transforms")
+                   callback_prob_factor,
+                   "Probability multiplier for running individual callback transforms")
         ->default_val(0.1)
-        ->check(CLI::Range(0.0, 1.0));
+        ->check(CLI::PositiveNumber);
+
+    constexpr std::array<std::string_view, 4> emulation_types
+        = {"none", "smoke", "regression", "extended"};
+    app.add_option("--emulation", "Run emulation tests")
+        ->check(CLI::IsMember(emulation_types))
+        ->each([&](const std::string& emulationtype) {
+            constexpr auto nidx = [emulation_types](const auto name) {
+                return std::find(emulation_types.begin(), emulation_types.end(), name)
+                       - emulation_types.begin();
+            };
+
+            // Emulation test suites focus on well-established software paths; we are looking for
+            // information about the hardware, which means that we aren't trying to find out a lot
+            // of information about the software.  Thus, no randomly-generated tests.
+            n_random_tests = 0;
+
+            // Run all of the emulation tests:
+            emulation_prob = 1.0;
+
+            // Callbacks are not an emulation test target.
+            callback_prob_factor = 0;
+
+            // We can do a switch on nidx(emulationtype) when we have C++20
+            // switch(nidx(emulationtype))
+            // {
+            // case nidx("smoke"):
+            // etc.
+
+            if(nidx(emulationtype) == nidx("smoke"))
+            {
+                // 2GB vram limit, approx 1 minute GPU time with short tests.
+                vramgb         = 2;
+                test_prob      = 0;
+                emulation_prob = 0.005;
+            }
+            if(nidx(emulationtype) == nidx("regression"))
+            {
+                vramgb         = 16;
+                emulation_prob = 1;
+                test_prob      = 0.01;
+            }
+            if(nidx(emulationtype) == nidx("extended"))
+            {
+                emulation_prob = 1;
+                test_prob      = 0.02;
+            }
+        });
+
     app.add_option("--fftw_compare", fftw_compare, "Compare to FFTW in accuracy tests")
         ->default_val(true);
     app.add_option("--mp_lib", mp_lib, "Multi-process library type: none (default), mpi")
@@ -301,38 +425,70 @@ int main(int argc, char* argv[])
             if(mp_lib == fft_params::fft_mp_lib_none)
             {
                 std::cout << "--mp_launch requires an mp library (see mp_lib in --help).\n";
-                std::exit(-1);
+                std::exit(EXIT_FAILURE);
             }
         })
         ->needs("--mp_lib");
 
-    CLI::Option* opt_seed
-        = app.add_option("--seed", random_seed, "Random seed; if unset, use an actual random seed");
+    app.add_flag("--smoketest", "Run a short (approx 5 minute) randomized selection of tests")
+        ->each([&](const std::string&) {
+            // The objective is to have an test that takes about 5 minutes, so just set the
+            // probability per test to a small value to achieve this result.
+            test_prob      = 0.001;
+            emulation_prob = 0.01;
+            n_random_tests = 10;
+        });
+
     app.add_flag("--callback", "Inject load/store callbacks")->each([&](const std::string&) {
         manual_params.run_callbacks = true;
     });
-    app.add_flag("--smoketest", "Run a short (approx 5 minute) randomized selection of tests")
-        ->each([&](const std::string&) {
-            // The objective is to have an test that takes about 5 minutes, so just set the probability
-            // per test to a small value to achieve this result.
-            test_prob = 0.002;
-        });
 
-    // Try parsing initial args that will be used to configure tests
-    // Allow extras to pass on gtest and rocFFT arguments without error
-    app.allow_extras();
-    try
     {
-        app.parse(argc, argv);
-    }
-    catch(const CLI::ParseError& e)
-    {
-        return app.exit(e);
+        // We explicitly scope opt_seed so that the object falls out of scope before the final
+        // parsing of the command line arguments.  Otherwise, the second parsing would mark the
+        // option as not having been specified, which can get rather confusing.
+
+        auto opt_seed = app.add_option(
+            "--seed", random_seed, "Random seed; if unset, use an actual random seed");
+
+        // Try parsing initial args that will be used to configure tests.
+        // Allow extras to pass on gtest and rocFFT arguments without error.
+        app.allow_extras();
+        try
+        {
+            app.parse(argc, argv);
+        }
+        catch(const CLI::ParseError& e)
+        {
+            return app.exit(e);
+        }
+
+        if(!*opt_seed)
+        {
+            std::cout << "Generating random seed: ";
+            std::random_device dev;
+            random_seed = dev();
+            std::cout << random_seed << std::endl;
+        }
     }
 
+    app.set_help_flag("");
+    auto opt_help = app.add_flag("-h, --help", "Produces this help message");
+
+    std::vector<std::string> remaining_args = app.remaining();
+    // Google test ignores the first element, so add something there so that it parses all of hte
+    // arguments that we want it to parse.:
+    remaining_args.insert(remaining_args.begin(), argv0);
     // NB: If we initialize gtest first, then it removes all of its own command-line
     // arguments and sets argc and argv correctly;
-    ::testing::InitGoogleTest(&argc, argv);
+    std::vector<char*> carg;
+    for(std::string& s : remaining_args)
+    {
+        carg.push_back(&s[0]);
+    }
+    carg.push_back(NULL);
+    decltype(argc) cargc = carg.size() - 1;
+    ::testing::InitGoogleTest(&cargc, carg.data());
 
     // Filename for fftw and fftwf wisdom.
     std::string fftw_wisdom_filename;
@@ -346,6 +502,9 @@ int main(int argc, char* argv[])
     // Full path to bitwise repro database file
     std::string repro_db_path;
 
+    // Bool option to just print tokens and exit
+    bool printtokens{false};
+
     // Declare the supported options. Some option pointers are declared to track passed opts.
     app.add_flag("--version", "Print queryable version information from the rocfft library")
         ->each([](const std::string&) {
@@ -358,7 +517,7 @@ int main(int argc, char* argv[])
     app.add_flag("--checkstride", "Check that data is not written outside of output strides")
         ->each([&](const std::string&) { manual_params.check_output_strides = true; });
 
-    CLI::Option* opt_token
+    auto opt_token
         = app.add_option("--token", test_token, "Test token name for manual test")->default_val("");
     // Group together options that conflict with --token
     auto* non_token = app.add_option_group("Token Conflict", "Options excluded by --token");
@@ -408,10 +567,7 @@ int main(int argc, char* argv[])
     app.add_option("--isize", manual_params.isize, "Logical size of input buffer");
     app.add_option("--osize", manual_params.osize, "Logical size of output buffer");
     app.add_option("--R", ramgb, "RAM limit in GiB for tests")
-        ->default_val(
-            (static_cast<size_t>(start_memory.total_bytes * system_memory::percentage_usable_memory)
-             + ONE_GiB - 1)
-            / ONE_GiB);
+        ->default_val(host_memory::singleton().get_total_gbytes());
     app.add_option("--V", vramgb, "VRAM limit in GiB for tests")->default_val(0);
     app.add_option("--half_epsilon", half_epsilon)->default_val(9.77e-4);
     app.add_option("--single_epsilon", single_epsilon)->default_val(3.75e-5);
@@ -437,11 +593,19 @@ int main(int argc, char* argv[])
                    precompile_file,
                    "Precompile kernels to a file for all test cases before running tests")
         ->default_val("");
+    app.add_flag("--printtokens", printtokens, "Print test tokens to scheduled to be run and exit");
+    // Default value is set in fft_params.h based on if device-side PRNG was enabled.
+    app.add_option("-g, --inputGen",
+                   manual_params.igen,
+                   "Input data generation:\n0) PRNG sequence (device)\n"
+                   "1) PRNG sequence (host)\n"
+                   "2) linearly-spaced sequence (device)\n"
+                   "3) linearly-spaced sequence (host)");
 
     // Parse rest of args and catch any errors here
     try
     {
-        app.parse(argc, argv);
+        app.parse(cargc, carg.data());
     }
     catch(const CLI::ParseError& e)
     {
@@ -455,30 +619,24 @@ int main(int argc, char* argv[])
     }
 
     // Ensure there are no leftover options used by neither gtest nor CLI11
-    std::vector<std::string> remaining_args = app.remaining();
-    if(!remaining_args.empty())
+    const auto leftover_args = app.remaining();
+    if(!leftover_args.empty())
     {
         std::cout << "Unrecognised option(s) found:\n  ";
-        for(auto i : app.remaining())
+        for(auto i : leftover_args)
             std::cout << i << " ";
         std::cout << "\nRun with --help for more information.\n";
         return EXIT_FAILURE;
     }
 
     std::cout << "half epsilon: " << half_epsilon << "\tsingle epsilon: " << single_epsilon
-              << "\tdouble epsilon: " << double_epsilon << "\n";
+              << "\tdouble epsilon: " << double_epsilon << std::endl;
+    std::cout << "Random seed: " << random_seed << std::endl;
 
-    if(!*opt_seed)
-    {
-        std::random_device dev;
-        random_seed = dev();
-    }
-    std::cout << "Random seed: " << random_seed << "\n";
-
-    // if precompiling, tell rocFFT to use the specified cache file
+    // If precompiling, tell rocFFT to use the specified cache file
     // to write kernels to
     //
-    // but if our environment already has a cache file for RTC, then
+    // But if our environment already has a cache file for RTC, then
     // we should just use that
     std::unique_ptr<EnvironmentSetTemp> env_precompile;
     if(!precompile_file.empty() && rocfft_getenv("ROCFFT_RTC_CACHE_PATH").empty())
@@ -488,9 +646,11 @@ int main(int argc, char* argv[])
     }
 
     rocfft_setup();
-    char v[256];
-    rocfft_get_version_string(v, 256);
-    std::cout << "rocFFT version: " << v << "\n";
+    {
+        char v[256];
+        rocfft_get_version_string(v, 256);
+        std::cout << "rocFFT version: " << v << std::endl;
+    }
 
 #ifdef FFTW_MULTITHREAD
     fftw_init_threads();
@@ -499,6 +659,9 @@ int main(int argc, char* argv[])
     fftwf_plan_with_nthreads(rocfft_concurrency());
 #endif
 
+    // Set host memory limit from command-line options
+    host_memory::singleton().set_limit_gbytes(ramgb);
+
     if(use_fftw_wisdom)
     {
         if(verbose)
@@ -550,7 +713,7 @@ int main(int argc, char* argv[])
 
     if(!test_token.empty())
     {
-        std::cout << "Reading fft params from token:\n" << test_token << "\n";
+        std::cout << "Reading fft params from token:\n" << test_token << std::endl;
 
         try
         {
@@ -558,8 +721,8 @@ int main(int argc, char* argv[])
         }
         catch(...)
         {
-            std::cout << "Unable to parse token.\n";
-            return 1;
+            std::cout << "Unable to parse token." << std::endl;
+            return EXIT_FAILURE;
         }
     }
     else
@@ -586,6 +749,17 @@ int main(int argc, char* argv[])
     if(!precompile_file.empty())
         precompile_test_kernels(precompile_file);
 
+    if(printtokens)
+    {
+        std::cout << "Tokens:" << std::endl;
+        const auto tokens = tokens_to_run();
+        for(const auto& token : tokens)
+        {
+            std::cout << token << std::endl;
+        }
+        return EXIT_SUCCESS;
+    }
+
     auto retval = RUN_ALL_TESTS();
 
     if(use_fftw_wisdom)
@@ -601,7 +775,14 @@ int main(int argc, char* argv[])
 
     rocfft_cleanup();
 
-    std::cout << "Random seed: " << random_seed << "\n";
+    const auto test_duration = std::chrono::system_clock::now() - test_begin;
+    const auto test_hours    = std::chrono::duration_cast<std::chrono::hours>(test_duration);
+    const auto test_minutes
+        = std::chrono::duration_cast<std::chrono::minutes>(test_duration - test_hours);
+    std::cout << "Test suite took " << test_hours.count() << " hours " << test_minutes.count()
+              << " minutes\n"
+              << std::endl;
+
     std::cout << "half precision max l-inf epsilon: " << max_linf_eps_half << "\n";
     std::cout << "half precision max l2 epsilon:     " << max_l2_eps_half << "\n";
     std::cout << "single precision max l-inf epsilon: " << max_linf_eps_single << "\n";
@@ -609,6 +790,12 @@ int main(int argc, char* argv[])
     std::cout << "double precision max l-inf epsilon: " << max_linf_eps_double << "\n";
     std::cout << "double precision max l2 epsilon:     " << max_l2_eps_double << "\n";
     std::cout << "Number of runtime issues: " << n_hip_failures << "\n";
+    std::cout << "Number of successful tests: "
+              << ::testing::UnitTest::GetInstance()->successful_test_count() << "\n";
+    std::cout << "Number of skipped tests: "
+              << ::testing::UnitTest::GetInstance()->skipped_test_count() << "\n";
+
+    std::cout << "\nRandom seed: " << random_seed << std::endl;
 
     return retval;
 }
@@ -623,8 +810,8 @@ TEST(manual, vs_fftw) // MANUAL TESTS HERE
         std::vector<unsigned int> deviceGrid(params.length.size() + 1, 1);
         deviceGrid[1] = manual_devices;
 
-        params.distribute_input(deviceGrid);
-        params.distribute_output(deviceGrid);
+        params.distribute_input(manual_devices, deviceGrid);
+        params.distribute_output(manual_devices, deviceGrid);
     }
 
     // Run an individual test using the provided command-line parameters.
@@ -648,13 +835,19 @@ TEST(manual, vs_fftw) // MANUAL TESTS HERE
     {
         GTEST_SKIP() << "host memory allocation failure";
     }
-    catch(ROCFFT_GTEST_SKIP& e)
+    catch(HOSTBUF_MEM_USAGE& e)
     {
-        GTEST_SKIP() << e.msg.str();
+        // explicitly clear test cache
+        last_cpu_fft_data = last_cpu_fft_cache();
+        GTEST_SKIP() << e.msg;
     }
-    catch(ROCFFT_GTEST_FAIL& e)
+    catch(ROCFFT_SKIP& e)
     {
-        GTEST_FAIL() << e.msg.str();
+        GTEST_SKIP() << e.msg;
+    }
+    catch(ROCFFT_FAIL& e)
+    {
+        GTEST_FAIL() << e.msg;
     }
 }
 
@@ -686,13 +879,13 @@ TEST(manual, bitwise_reproducibility) // MANUAL TESTS HERE
     {
         GTEST_SKIP() << "host memory allocation failure";
     }
-    catch(ROCFFT_GTEST_SKIP& e)
+    catch(ROCFFT_SKIP& e)
     {
-        GTEST_SKIP() << e.msg.str();
+        GTEST_SKIP() << e.msg;
     }
-    catch(ROCFFT_GTEST_FAIL& e)
+    catch(ROCFFT_FAIL& e)
     {
-        GTEST_FAIL() << e.msg.str();
+        GTEST_FAIL() << e.msg;
     }
     SUCCEED();
 }
