@@ -23,10 +23,12 @@
 #include "../../shared/concurrency.h"
 #include "../../shared/environment.h"
 #include "../../shared/gpubuf.h"
+#include "../../shared/precision_type.h"
 #include "../../shared/rocfft_complex.h"
 #include "hip/hip_runtime_api.h"
 #include <boost/scope_exit.hpp>
 #include <condition_variable>
+#include <cstdio>
 #include <cstdlib>
 #include <fstream>
 #include <gtest/gtest.h>
@@ -101,6 +103,167 @@ TEST(rocfft_UnitTest, plan_description)
     ASSERT_TRUE(rocfft_status_success == rocfft_plan_destroy(plan));
 }
 
+TEST(rocfft_UnitTest, plan_description_reuse)
+{
+    // check that a plan description can be reused between different
+    // plans, with different layout parameters for each.
+
+    // allocate plan description once
+    rocfft_plan_description desc = nullptr;
+    ASSERT_EQ(rocfft_plan_description_create(&desc), rocfft_status_success);
+
+    std::vector<rocfft_complex<float>> output;
+
+    // do length-8 FFTs with different strides.  first one is
+    // stride-1 and we use that as our baseline to know what output
+    // to expect for the rest
+    const size_t length = 8;
+    for(const size_t stride : {1, 2, 4})
+    {
+        // set layout for this stride
+        ASSERT_EQ(rocfft_plan_description_set_data_layout(desc,
+                                                          rocfft_array_type_complex_interleaved,
+                                                          rocfft_array_type_complex_interleaved,
+                                                          nullptr,
+                                                          nullptr,
+                                                          1,
+                                                          &stride,
+                                                          length * stride,
+                                                          1,
+                                                          &stride,
+                                                          length * stride),
+                  rocfft_status_success);
+
+        static const rocfft_complex<float> input[8]{{-0.100, 0.380},
+                                                    {0.0166, 0.439},
+                                                    {-0.475, 0.212},
+                                                    {0.440, -0.432},
+                                                    {0.445, 0.0589},
+                                                    {0.296, 0.164},
+                                                    {-0.084, 0.077},
+                                                    {0.320, 0.087}};
+
+        // allocate host buffer.  initialize the whole thing to zero
+        // but set a known input along the strides we want
+        std::vector<rocfft_complex<float>> data_host(length * stride, {0.0, 0.0});
+        for(size_t i = 0; i < length; ++i)
+        {
+            data_host[i * stride] = input[i];
+        }
+
+        // copy to device
+        const size_t data_bytes = data_host.size() * sizeof(rocfft_complex<float>);
+        gpubuf_t<rocfft_complex<float>> data_dev;
+        ASSERT_EQ(data_dev.alloc(data_bytes), hipSuccess);
+        void* data_dev_ptr = data_dev.data();
+        ASSERT_EQ(hipMemcpy(data_dev_ptr, data_host.data(), data_bytes, hipMemcpyHostToDevice),
+                  hipSuccess);
+
+        // do the transform
+        rocfft_plan plan = nullptr;
+        ASSERT_EQ(rocfft_plan_create(&plan,
+                                     rocfft_placement_inplace,
+                                     rocfft_transform_type_complex_forward,
+                                     rocfft_precision_single,
+                                     1,
+                                     &length,
+                                     1,
+                                     desc),
+                  rocfft_status_success);
+        ASSERT_EQ(rocfft_execute(plan, &data_dev_ptr, nullptr, nullptr), rocfft_status_success);
+        ASSERT_EQ(hipMemcpy(data_host.data(), data_dev_ptr, data_bytes, hipMemcpyDeviceToHost),
+                  hipSuccess);
+        ASSERT_EQ(hipDeviceSynchronize(), hipSuccess);
+
+        // save output for reference on first run
+        if(output.empty())
+        {
+            output = data_host;
+        }
+        else
+        {
+            // check that the output matches output from the first
+            // (stride-1) run.
+            for(size_t i = 0; i < length; ++i)
+                ASSERT_EQ(data_host[i * stride], output[i]);
+        }
+        ASSERT_EQ(rocfft_plan_destroy(plan), rocfft_status_success);
+    }
+
+    ASSERT_EQ(rocfft_plan_description_destroy(desc), rocfft_status_success);
+}
+
+// run a transform with all log levels enabled
+TEST(rocfft_UnitTest, log_levels)
+{
+    // clean up environment and temporary file when we exit
+    BOOST_SCOPE_EXIT_ALL(=)
+    {
+        rocfft_cleanup();
+        // re-init logs with default logging
+        rocfft_setup();
+    };
+    rocfft_cleanup();
+
+    // enumerate all known log levels and direct all of the logs to nowhere
+    EnvironmentSetTemp layer("ROCFFT_LAYER", std::to_string(0xffffffff).c_str());
+#ifdef WIN32
+    static const char* log_output = "NUL";
+#else
+    static const char* log_output   = "/dev/null";
+#endif
+    EnvironmentSetTemp log_trace_path("ROCFFT_LOG_TRACE_PATH", log_output);
+    EnvironmentSetTemp log_bench_path("ROCFFT_LOG_BENCH_PATH", log_output);
+    EnvironmentSetTemp log_profile_path("ROCFFT_LOG_PROFILE_PATH", log_output);
+    EnvironmentSetTemp log_plan_path("ROCFFT_LOG_PLAN_PATH", log_output);
+    EnvironmentSetTemp log_kernelio_path("ROCFFT_LOG_KERNELIO_PATH", log_output);
+    EnvironmentSetTemp log_rtc_path("ROCFFT_LOG_RTC_PATH", log_output);
+    EnvironmentSetTemp log_tuning_path("ROCFFT_LOG_TUNING_PATH", log_output);
+    EnvironmentSetTemp log_graph_path("ROCFFT_LOG_GRAPH_PATH", log_output);
+
+    rocfft_setup();
+
+    // Test single-kernel Bluestein and a multi-kernel plan
+    //
+    // TODO: add fused L1D Bluestein case like 8191, as that does weird
+    // things with buffers
+    for(const size_t length : {
+            37,
+            64,
+            32768,
+        })
+    {
+        for(const auto type : {rocfft_transform_type_complex_forward,
+                               rocfft_transform_type_real_forward,
+                               rocfft_transform_type_real_inverse})
+        {
+            for(const auto precision :
+                {rocfft_precision_single, rocfft_precision_double, rocfft_precision_half})
+            {
+                rocfft_plan plan = nullptr;
+                ASSERT_EQ(
+                    rocfft_plan_create(
+                        &plan, rocfft_placement_inplace, type, precision, 1, &length, 1, nullptr),
+                    rocfft_status_success);
+
+                // assume transform uses complex, will overallocate for real
+                // transforms but we only care about logging
+                gpubuf data_dev;
+                ASSERT_EQ(
+                    data_dev.alloc(element_size(precision, rocfft_array_type_complex_interleaved)
+                                   * length),
+                    hipSuccess);
+
+                void* data_dev_ptr = data_dev.data();
+                ASSERT_EQ(rocfft_execute(plan, &data_dev_ptr, nullptr, nullptr),
+                          rocfft_status_success);
+
+                rocfft_plan_destroy(plan);
+            }
+        }
+    }
+}
+
 // Check whether logs can be emitted from multiple threads properly
 TEST(rocfft_UnitTest, log_multithreading)
 {
@@ -507,8 +670,17 @@ TEST(rocfft_UnitTest, rtc_test_harness)
 
     // extra scope to control lifetime of env vars
     {
-        // rtc test harness writes to system's temp directory
-        auto tmp_path = fs::temp_directory_path();
+        // create a temporary directory to hold all of the temp files
+        // that get created
+        const fs::path tmp_path = std::tmpnam(nullptr);
+        try
+        {
+            fs::create_directory(tmp_path);
+        }
+        catch(fs::filesystem_error& e)
+        {
+            GTEST_SKIP() << "unable to create temp dir for test harnesses: " << e.what();
+        }
 
         // activate writing of rtc test harnesses
         EnvironmentSetTemp env_harness("ROCFFT_DEBUG_GENERATE_KERNEL_HARNESS", "1");
@@ -523,17 +695,6 @@ TEST(rocfft_UnitTest, rtc_test_harness)
 
         rocfft_setup();
 
-        // ensure stale files from previous runs of this test won't cause
-        // problems - clean up any rocfft_kernel_harness_*.cpp files that
-        // might be left behind
-        for(const auto& entry : std::filesystem::directory_iterator{tmp_path})
-        {
-            auto filename = entry.path().filename();
-            if(filename.string().compare(0, 22, "rocfft_kernel_harness_") == 0
-               && filename.extension().string() == ".cpp")
-                fs::remove(entry);
-        }
-
         // construct a few different types of plans to try to get all
         // different kernels compiled
 
@@ -614,6 +775,19 @@ TEST(rocfft_UnitTest, rtc_test_harness)
 
         // check that all compiles succeeded
         for(const auto& file : files)
+        {
             ASSERT_EQ(file.second, 0);
+        }
+
+        // clean up temporary files
+        try
+        {
+            fs::remove_all(tmp_path);
+        }
+        catch(fs::filesystem_error&)
+        {
+            // this should work, but ignore errors as the build
+            // status is what matters for this test
+        }
     }
 }
