@@ -278,14 +278,7 @@ bool AssignmentPolicy::ValidOutBuffer(ExecPlan&           execPlan,
         if(BufferIsUnitStride(execPlan, buffer))
         {
             // just check if there's enough space
-            return std::accumulate(nodeLen.begin(),
-                                   nodeLen.end(),
-                                   static_cast<size_t>(1),
-                                   std::multiplies<size_t>())
-                   <= std::accumulate(bufLen.begin(),
-                                      bufLen.end(),
-                                      static_cast<size_t>(1),
-                                      std::multiplies<size_t>());
+            return product(nodeLen.begin(), nodeLen.end()) <= product(bufLen.begin(), bufLen.end());
         }
 
         // ensure that the node's dimensions fit exactly into the
@@ -551,7 +544,10 @@ void AssignmentPolicy::AssignChirpBuffers(ExecPlan& execPlan)
         chirpFwdNodePlan.length    = chirpFwdNode->length;
         chirpFwdNodePlan.dimension = chirpFwdNode->dimension;
 
-        ExecPlan execPlanFwdChirp;
+        // create a temporary plan that just does chirp - the plan nodes
+        // get copied into our actual plan so the comm_rank, mgpuPlan,
+        // location don't need to be correct
+        ExecPlan execPlanFwdChirp{0, false, {}};
         execPlanFwdChirp.IsChirpPlan = true;
         execPlanFwdChirp.rootPlan    = NodeFactory::CreateExplicitNode(chirpFwdNodePlan, nullptr);
 
@@ -1238,6 +1234,10 @@ void AssignmentPolicy::PadPlan(ExecPlan& execPlan)
         return;
 
     RecursiveTraverse(execPlan.rootPlan.get(), [&execPlan](TreeNode* n) {
+        // Skip nodes that have partial passes
+        if(n->applyPartialPass)
+            return;
+
         // Look for nodes that begin writing to a new temp buffer
         // (i.e. obOut is a paddable temp buffer, and obIn was a
         // different buffer)
