@@ -1,4 +1,4 @@
-// Copyright (C) 2016 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2016 - 2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -32,11 +32,13 @@
 #include "../../../shared/gpubuf.h"
 #include "../../../shared/hip_object_wrapper.h"
 #include "../../../shared/rocfft_complex.h"
+#include "../../shared/device_properties.h"
 #include "../device/kernels/callback.h"
 #include "../device/kernels/common.h"
 #include "compute_scheme.h"
 #include "enum_printer.h"
 #include "function_map_key.h"
+#include "function_pool.h"
 #include "kargs.h"
 #include "load_store_ops.h"
 #include "rtc_kernel.h"
@@ -60,8 +62,6 @@ enum FuseType
     FT_STOCKHAM_R2C_TRANSPOSE, // Stokham + post-r2c + transpose (Advance of FT_R2C_TRANSPOSE)
 };
 
-typedef void (*DevFnCall)(const void*, void*);
-
 struct GridParam
 {
     unsigned int b_x, b_y, b_z; // in HIP, the data type of dimensions of work
@@ -89,8 +89,6 @@ static std::string get_arch_name(const hipDeviceProp_t& prop)
                                                        "gfx906",
                                                        "gfx908",
                                                        "gfx90a",
-                                                       "gfx940",
-                                                       "gfx941",
                                                        "gfx942",
                                                        "gfx1030",
                                                        "gfx1100",
@@ -166,6 +164,7 @@ static SchemeVec EmptySchemeVec = {};
 
 class TreeNode;
 class LeafNode;
+class function_pool;
 
 // The mininal tree node data needed to decide the scheme
 struct NodeMetaData
@@ -260,13 +259,14 @@ class TreeNode
 protected:
     TreeNode(TreeNode* p)
         : parent(p)
+        , deviceProp(p ? p->deviceProp : get_curr_device_prop())
+        , pool(deviceProp)
     {
         if(p != nullptr)
         {
-            precision  = p->precision;
-            batch      = p->batch;
-            direction  = p->direction;
-            deviceProp = p->deviceProp;
+            precision = p->precision;
+            batch     = p->batch;
+            direction = p->direction;
         }
 
         allowedOutBuf
@@ -319,9 +319,6 @@ public:
     // Direction of the transform (-1: forward, +1: inverse)
     int direction = -1;
 
-    // The number of padding at the end of each row in lds
-    unsigned int lds_padding = 0;
-
     // Data format parameters:
     rocfft_result_placement placement    = rocfft_placement_inplace;
     rocfft_precision        precision    = rocfft_precision_single;
@@ -400,6 +397,7 @@ public:
     UserCallbacks callbacks;
 
     hipDeviceProp_t deviceProp = {};
+    function_pool   pool;
 
     // comments inserted by optimization passes to explain changes done
     // to the node
@@ -561,7 +559,7 @@ public:
     virtual bool KernelCheck(std::vector<FMKey>& kernel_keys = EmptyFMKeyVec) = 0;
     virtual bool CreateDevKernelArgs()                                        = 0;
     virtual bool CreateDeviceResources()                                      = 0;
-    virtual void SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)    = 0;
+    virtual void SetupGridParam(GridParam& gp)                                = 0;
 
     // for 3D SBRC kernels, decide the transpose type based on the
     // block width and lengths that the block tiles need to align on.
@@ -622,9 +620,9 @@ protected:
         return false;
     }
 
-    void SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp) override
+    void SetupGridParam(GridParam& gp) override
     {
-        throw std::runtime_error("Shouldn't call SetupGridParamAndFuncPtr in a non-LeafNode");
+        throw std::runtime_error("Shouldn't call SetupGridParam in a non-LeafNode");
     }
 
 public:
@@ -669,7 +667,7 @@ public:
     {
         return 0;
     }
-    virtual void SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp) = 0;
+    virtual void SetupGridParam_internal(GridParam& gp) = 0;
 
 public:
     // leaf node would print additional informations about kernel setting
@@ -679,7 +677,7 @@ public:
                              std::vector<FMKey>& kernel_keys     = EmptyFMKeyVec) override;
     virtual bool CreateDevKernelArgs() override;
     bool         CreateDeviceResources() override;
-    void         SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp) override;
+    void         SetupGridParam(GridParam& gp) override;
     FMKey        GetKernelKey() const override;
     virtual void GetKernelFactors();
     virtual void GetKernelPartialPassFactors();
@@ -701,7 +699,7 @@ protected:
         allowInplace = false;
     }
 
-    void SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp) override;
+    void SetupGridParam_internal(GridParam& gp) override;
 
 public:
     // Transpose tiles read more row-ish and write more column-ish.  So
@@ -1013,19 +1011,35 @@ struct MultiPlanItem
 // Communication operations
 struct CommPointToPoint : public MultiPlanItem
 {
-    rocfft_precision  precision;
-    rocfft_array_type arrayType;
-
-    // number of elements to copy
-    size_t numElems;
-
-    rocfft_location_t srcLocation;
-    BufferPtr         srcPtr;
-    size_t            srcOffset = 0;
-
-    rocfft_location_t destLocation;
-    BufferPtr         destPtr;
-    size_t            destOffset = 0;
+    CommPointToPoint(int               local_comm_rank,
+                     rocfft_precision  _precision,
+                     rocfft_array_type _arrayType,
+                     size_t            _numElems,
+                     rocfft_location_t _srcLocation,
+                     BufferPtr         _srcPtr,
+                     size_t            _srcOffset,
+                     rocfft_location_t _destLocation,
+                     BufferPtr         _destPtr,
+                     size_t            _destOffset)
+        : precision(_precision)
+        , arrayType(_arrayType)
+        , numElems(_numElems)
+        , srcLocation(_srcLocation)
+        , srcPtr(_srcPtr)
+        , srcOffset(_srcOffset)
+        , destLocation(_destLocation)
+        , destPtr(_destPtr)
+        , destOffset(_destOffset)
+    {
+        // same-rank communication needs a stream and event to signal completion
+        if(local_comm_rank == srcLocation.comm_rank
+           && srcLocation.comm_rank == destLocation.comm_rank)
+        {
+            rocfft_scoped_device dev(srcLocation.device);
+            stream.alloc();
+            event.alloc();
+        }
+    }
 
     void ExecuteAsync(const rocfft_plan     plan,
                       void*                 in_buffer[],
@@ -1047,6 +1061,19 @@ struct CommPointToPoint : public MultiPlanItem
     }
 
 private:
+    const rocfft_precision  precision;
+    const rocfft_array_type arrayType;
+
+    // number of elements to copy
+    const size_t numElems;
+
+    const rocfft_location_t srcLocation;
+    const BufferPtr         srcPtr;
+    const size_t            srcOffset = 0;
+
+    const rocfft_location_t destLocation;
+    const BufferPtr         destPtr;
+    const size_t            destOffset = 0;
     // Stream to run the async operation in
     hipStream_wrapper_t stream;
     // Event to signal when the async operations are finished.
@@ -1057,11 +1084,17 @@ private:
 // create an MPI group with those ranks.
 struct CommScatter : public MultiPlanItem
 {
-    rocfft_precision  precision;
-    rocfft_array_type arrayType;
+    CommScatter(rocfft_precision  _precision,
+                rocfft_array_type _arrayType,
 
-    rocfft_location_t srcLocation;
-    BufferPtr         srcPtr;
+                rocfft_location_t _srcLocation,
+                BufferPtr         _srcPtr)
+        : precision(_precision)
+        , arrayType(_arrayType)
+        , srcLocation(_srcLocation)
+        , srcPtr(_srcPtr)
+    {
+    }
 
     // one or more ranks to send data to
     struct ScatterOp
@@ -1088,7 +1121,22 @@ struct CommScatter : public MultiPlanItem
         // Number of elements to copy
         size_t numElems;
     };
-    std::vector<ScatterOp> ops;
+
+    // add an operation to scatter to a destination
+    void AddOperation(int local_comm_rank, ScatterOp&& op)
+    {
+        // same-rank communication needs a stream and event to signal completion
+        if(srcLocation.comm_rank == local_comm_rank
+           && op.destLocation.comm_rank == srcLocation.comm_rank)
+        {
+            rocfft_scoped_device dev(srcLocation.device);
+            if(!stream)
+                stream.alloc();
+            if(!event)
+                event.alloc();
+        }
+        ops.emplace_back(std::move(op));
+    }
 
     void ExecuteAsync(const rocfft_plan     plan,
                       void*                 in_buffer[],
@@ -1118,6 +1166,13 @@ struct CommScatter : public MultiPlanItem
     }
 
 private:
+    const rocfft_precision  precision;
+    const rocfft_array_type arrayType;
+    const rocfft_location_t srcLocation;
+    const BufferPtr         srcPtr;
+
+    std::vector<ScatterOp> ops;
+
     // Stream to run the async operations in
     hipStream_wrapper_t stream;
     // Event to signal when the async operations are finished.
@@ -1128,11 +1183,17 @@ private:
 // create an MPI group with those ranks.
 struct CommGather : public MultiPlanItem
 {
-    rocfft_precision  precision;
-    rocfft_array_type arrayType;
-
-    rocfft_location_t destLocation;
-    BufferPtr         destPtr;
+    CommGather(int               local_comm_rank,
+               rocfft_precision  _precision,
+               rocfft_array_type _arrayType,
+               rocfft_location_t _destLocation,
+               BufferPtr         _destPtr)
+        : precision(_precision)
+        , arrayType(_arrayType)
+        , destLocation(_destLocation)
+        , destPtr(_destPtr)
+    {
+    }
 
     // one or more ranks to get data from
     struct GatherOp
@@ -1159,7 +1220,27 @@ struct CommGather : public MultiPlanItem
         // Number of elements to copy
         size_t numElems;
     };
-    std::vector<GatherOp> ops;
+
+    // add an operation to gather from a source
+    void AddOperation(int local_comm_rank, GatherOp&& op)
+    {
+        // allocate space to allow each gather operation to have a
+        // stream/event, though we only allocate streams/events for
+        // rank-local communication
+        streams.resize(streams.size() + 1);
+        events.resize(events.size() + 1);
+
+        // look at the actual operation and allocate stream/event for
+        // rank-local communication, so we can get completion information
+        if(destLocation.comm_rank == op.srcLocation.comm_rank
+           && op.srcLocation.comm_rank == local_comm_rank)
+        {
+            rocfft_scoped_device dev(op.srcLocation.device);
+            streams.back().alloc();
+            events.back().alloc();
+        }
+        ops.emplace_back(std::move(op));
+    }
 
     void ExecuteAsync(const rocfft_plan     plan,
                       void*                 in_buffer[],
@@ -1183,6 +1264,15 @@ struct CommGather : public MultiPlanItem
                   });
     }
 
+private:
+    const rocfft_precision  precision;
+    const rocfft_array_type arrayType;
+
+    const rocfft_location_t destLocation;
+    const BufferPtr         destPtr;
+
+    std::vector<GatherOp> ops;
+
     // Streams to run the async operations in - since each memcpy is
     // coming from a different source device, each needs a separate
     // stream
@@ -1194,29 +1284,58 @@ struct CommGather : public MultiPlanItem
 // Send data from all ranks to all ranks in the plan.  Each rank must
 // send from/to a single buffer (with different read/write offsets
 // for each other rank).
-struct CommAllToAllv : public MultiPlanItem
+// The all-to-all communication is performed using MPI_Ialltoall or MPI_Ialltoallv.
+// The former is preferable, as it is usually more optimized.
+struct CommAllToAll : public MultiPlanItem
 {
-    CommAllToAllv() = default;
+    CommAllToAll(rocfft_precision           _precision,
+                 rocfft_array_type          _arrayType,
+                 const std::vector<size_t>& _sendOffsets,
+                 const std::vector<size_t>& _sendCounts,
+                 const std::vector<size_t>& _recvOffsets,
+                 const std::vector<size_t>& _recvCounts,
+                 BufferPtr                  _sendBuf,
+                 BufferPtr                  _recvBuf)
+        : precision(_precision)
+        , arrayType(_arrayType)
+        , sendOffsets(_sendOffsets)
+        , sendCounts(_sendCounts)
+        , recvOffsets(_recvOffsets)
+        , recvCounts(_recvCounts)
+        , sendBuf(_sendBuf)
+        , recvBuf(_recvBuf)
+    {
+        // Currently MPI interface uses 32-bit signed ints, so assert
+        // that our counts/offsets don't overflow that type
+        auto checkArray = [](const std::vector<size_t>& arr) {
+            if(std::any_of(arr.begin(), arr.end(), [](size_t elem) {
+                   return elem > std::numeric_limits<int>::max();
+               }))
+                throw std::runtime_error("count/offset exceeds int max");
+        };
 
-    rocfft_precision  precision;
-    rocfft_array_type arrayType;
+        checkArray(sendOffsets);
+        checkArray(sendCounts);
+        checkArray(recvOffsets);
+        checkArray(recvCounts);
+    }
 
-    // counts and offsets are all in elements (where element size is
-    // knowable from precision + array type), for the current rank
-    std::vector<size_t> sendOffsets;
-    std::vector<size_t> sendCounts;
-    std::vector<size_t> recvOffsets;
-    std::vector<size_t> recvCounts;
-
-    // send/receive buffers
-    BufferPtr sendBuf;
-    BufferPtr recvBuf;
+    // enum for error handling for different multi-process communication
+    // libraries such as MPI, RCCL, rocSHMEM, etc.
+    enum CommStatus
+    {
+        COMM_SUCCESS,
+        COMM_MPI_ERROR
+    };
+    CommStatus  comm_status = COMM_SUCCESS;
+    std::string error_message;
 
     void ExecuteAsync(const rocfft_plan     plan,
                       void*                 in_buffer[],
                       void*                 out_buffer[],
                       rocfft_execution_info info,
                       size_t                multiPlanIdx) override;
+
     void Wait() override;
 
     void Print(rocfft_ostream& os, const int indent) const override;
@@ -1232,6 +1351,21 @@ struct CommAllToAllv : public MultiPlanItem
         // runs on all ranks
         return true;
     }
+
+private:
+    const rocfft_precision  precision;
+    const rocfft_array_type arrayType;
+
+    // counts and offsets are all in elements (where element size is
+    // knowable from precision + array type), for the current rank
+    const std::vector<size_t> sendOffsets;
+    const std::vector<size_t> sendCounts;
+    const std::vector<size_t> recvOffsets;
+    const std::vector<size_t> recvCounts;
+
+    // send/receive buffers
+    const BufferPtr sendBuf;
+    const BufferPtr recvBuf;
 };
 
 // Tree-structured FFT plan.  This is specific to a single device on
@@ -1239,6 +1373,23 @@ struct CommAllToAllv : public MultiPlanItem
 // memory allocated for things like kernel arguments and twiddles.
 struct ExecPlan : public MultiPlanItem
 {
+    ExecPlan(int local_comm_rank, bool _mgpuPlan, rocfft_location_t _location)
+        : location(_location)
+        , mgpuPlan(_mgpuPlan)
+    {
+        // in a multi-GPU plan, this plan is just one piece of the overall
+        // distributed FFT.  need to allocate stream/event for completion
+        // signalling
+        if(mgpuPlan && location.comm_rank == local_comm_rank)
+        {
+            // if the user provides a stream for us at execute time, we
+            // might not use this stream.  but allocate it during plan
+            // creation because it can be slow to do it during execution
+            rocfft_scoped_device dev(location.device);
+            stream.alloc();
+            event.alloc();
+        }
+    }
     // device where this work will be executed
     rocfft_location_t location;
 
@@ -1285,7 +1436,6 @@ struct ExecPlan : public MultiPlanItem
     // flattened potentially-fusable shims of rootPlan
     std::vector<FuseShim*> fuseShims;
 
-    std::vector<DevFnCall> devFnCall;
     std::vector<GridParam> gridParam;
 
     hipDeviceProp_t deviceProp;
