@@ -70,12 +70,11 @@ bool PlanPowX(ExecPlan& execPlan)
 
     for(const auto& node : execPlan.execSeq)
     {
-        DevFnCall ptr = nullptr;
         GridParam gp;
 
         // if this kernel is runtime compiled, use the grid params
         // from compilation as a default.  the node is free to
-        // override this default in its SetupGPAndFnPtr_internal
+        // override this default in its SetupGridParam_internal
         // method.
         auto& rtcKernel = node->compiledKernel.get();
         if(rtcKernel)
@@ -87,9 +86,8 @@ bool PlanPowX(ExecPlan& execPlan)
             gp.wgs_y = rtcKernel->blockDim.y;
             gp.wgs_z = rtcKernel->blockDim.z;
         }
-        node->SetupGridParamAndFuncPtr(ptr, gp);
+        node->SetupGridParam(gp);
 
-        execPlan.devFnCall.push_back(ptr);
         execPlan.gridParam.push_back(gp);
     }
 
@@ -187,8 +185,7 @@ static size_t data_size_bytes(const std::vector<size_t>& lengths,
                               rocfft_array_type          type)
 {
     // first compute the raw number of elements
-    const size_t elems = std::accumulate(
-        lengths.begin(), lengths.end(), static_cast<size_t>(1), std::multiplies<size_t>());
+    const size_t elems = product(lengths.begin(), lengths.end());
     // size of each element
     const size_t elemsize = real_type_size(precision);
     switch(type)
@@ -337,20 +334,20 @@ void DebugPrintBuffer(rocfft_ostream&             stream,
         {
         case rocfft_precision_half:
         {
-            buffer_printer<_Float16> s;
-            s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+            buffer_printer<rocfft_fp16> s;
+            s.print_buffer_half(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
             break;
         }
         case rocfft_precision_single:
         {
             buffer_printer<float> s;
-            s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+            s.print_buffer_single(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
             break;
         }
         case rocfft_precision_double:
         {
             buffer_printer<double> s;
-            s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+            s.print_buffer_double(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
             break;
         }
         }
@@ -366,14 +363,16 @@ void DebugPrintBuffer(rocfft_ostream&             stream,
             case rocfft_array_type_complex_interleaved:
             case rocfft_array_type_hermitian_interleaved:
             {
-                buffer_printer<rocfft_complex<_Float16>> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                buffer_printer<rocfft_complex<rocfft_fp16>> s;
+                s.print_buffer_half(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             case rocfft_array_type_real:
             {
-                buffer_printer<_Float16> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                buffer_printer<rocfft_fp16> s;
+                s.print_buffer_half(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             default:
@@ -389,13 +388,15 @@ void DebugPrintBuffer(rocfft_ostream&             stream,
             case rocfft_array_type_hermitian_interleaved:
             {
                 buffer_printer<rocfft_complex<float>> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                s.print_buffer_single(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             case rocfft_array_type_real:
             {
                 buffer_printer<float> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                s.print_buffer_single(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             default:
@@ -411,13 +412,15 @@ void DebugPrintBuffer(rocfft_ostream&             stream,
             case rocfft_array_type_hermitian_interleaved:
             {
                 buffer_printer<rocfft_complex<double>> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                s.print_buffer_double(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             case rocfft_array_type_real:
             {
                 buffer_printer<double> s;
-                s.print_buffer(bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
+                s.print_buffer_double(
+                    bufvec, length_rm, stride_rm, batch, dist, print_offset, stream);
                 break;
             }
             default:
@@ -528,7 +531,6 @@ void TransformPowX(const ExecPlan&       execPlan,
                    rocfft_execution_info info,
                    size_t                multiPlanIdx)
 {
-    assert(execPlan.execSeq.size() == execPlan.devFnCall.size());
     assert(execPlan.execSeq.size() == execPlan.gridParam.size());
 
     bool processing_tuning = TuningBenchmarker::GetSingleton().IsProcessingTuning();
@@ -568,10 +570,6 @@ void TransformPowX(const ExecPlan&       execPlan,
         data.node          = execPlan.execSeq[i];
         data.rocfft_stream = (info == nullptr) ? 0 : info->rocfft_stream;
         data.deviceProp    = execPlan.deviceProp;
-        if(LOG_PLAN_ENABLED())
-            data.log_func = log_plan;
-        else
-            data.log_func = nullptr;
 
         // Size of complex type
         const size_t complexTSize = complex_type_size(data.node->precision);
@@ -779,118 +777,107 @@ void TransformPowX(const ExecPlan&       execPlan,
             *kernelio_stream << std::endl;
         }
 
-        DevFnCall fn = execPlan.devFnCall[i];
-        if(fn || data.node->compiledKernel.get())
+#ifdef REF_DEBUG
+        rocfft_cout << "\n---------------------------------------------\n";
+        rocfft_cout << "\n\nkernel: " << i << std::endl;
+        rocfft_cout << "\tscheme: " << PrintScheme(execPlan.execSeq[i]->scheme) << std::endl;
+        rocfft_cout << "\titype: " << execPlan.execSeq[i]->inArrayType << std::endl;
+        rocfft_cout << "\totype: " << execPlan.execSeq[i]->outArrayType << std::endl;
+        rocfft_cout << "\tlength: ";
+        for(const auto& i : execPlan.execSeq[i]->length)
         {
-#ifdef REF_DEBUG
-            rocfft_cout << "\n---------------------------------------------\n";
-            rocfft_cout << "\n\nkernel: " << i << std::endl;
-            rocfft_cout << "\tscheme: " << PrintScheme(execPlan.execSeq[i]->scheme) << std::endl;
-            rocfft_cout << "\titype: " << execPlan.execSeq[i]->inArrayType << std::endl;
-            rocfft_cout << "\totype: " << execPlan.execSeq[i]->outArrayType << std::endl;
-            rocfft_cout << "\tlength: ";
-            for(const auto& i : execPlan.execSeq[i]->length)
-            {
-                rocfft_cout << i << " ";
-            }
-            rocfft_cout << std::endl;
-            rocfft_cout << "\tbatch:   " << execPlan.execSeq[i]->batch << std::endl;
-            rocfft_cout << "\tidist:   " << execPlan.execSeq[i]->iDist << std::endl;
-            rocfft_cout << "\todist:   " << execPlan.execSeq[i]->oDist << std::endl;
-            rocfft_cout << "\tistride:";
-            for(const auto& i : execPlan.execSeq[i]->inStride)
-            {
-                rocfft_cout << " " << i;
-            }
-            rocfft_cout << std::endl;
-            rocfft_cout << "\tostride:";
-            for(const auto& i : execPlan.execSeq[i]->outStride)
-            {
-                rocfft_cout << " " << i;
-            }
-            rocfft_cout << std::endl;
-
-            RefLibOp refLibOp(&data);
-#endif
-
-            // execution kernel:
-            if(emit_profile_log)
-                if(hipEventRecord(start) != hipSuccess)
-                    throw std::runtime_error("hipEventRecord failure");
-
-            DeviceCallOut back;
-
-            // give callback parameters to kernel launcher
-            data.callbacks = execPlan.execSeq[i]->callbacks;
-
-            // choose which compiled kernel to run
-            RTCKernel* localCompiledKernel
-                = data.get_callback_type() == CallbackType::NONE
-                      ? data.node->compiledKernel.get().get()
-                      : data.node->compiledKernelWithCallbacks.get().get();
-
-            if(localCompiledKernel)
-                localCompiledKernel->launch(data, data.node->deviceProp);
-            else
-                fn(&data, &back);
-
-            if(emit_profile_log)
-                if(hipEventRecord(stop) != hipSuccess)
-                    throw std::runtime_error("hipEventRecord failure");
-
-            // If we were on the null stream, measure elapsed time
-            // and emit profile logging.  If a stream was given, we
-            // can't wait for the transform to finish, so we can't
-            // emit any information.
-            if(emit_profile_log)
-            {
-                if(hipEventSynchronize(stop) != hipSuccess)
-                    throw std::runtime_error("hipEventSynchronize failure");
-                size_t in_size_bytes = data_size_bytes(
-                    data.node->length, data.node->precision, data.node->inArrayType);
-                size_t out_size_bytes = data_size_bytes(
-                    data.node->length, data.node->precision, data.node->outArrayType);
-                size_t total_size_bytes = (in_size_bytes + out_size_bytes) * data.node->batch;
-
-                float duration_ms = 0.0f;
-                if(hipEventElapsedTime(&duration_ms, start, stop) != hipSuccess)
-                    throw std::runtime_error("hipEventElapsedTime failure");
-                auto exec_bw        = execution_bandwidth_GB_per_s(total_size_bytes, duration_ms);
-                auto efficiency_pct = 0.0;
-                if(max_memory_bw != 0.0)
-                    efficiency_pct = 100.0 * exec_bw / max_memory_bw;
-                if(processing_tuning)
-                    tuningPacket->bw_effs[i] = efficiency_pct;
-
-                log_profile(__func__,
-                            "scheme",
-                            PrintScheme(execPlan.execSeq[i]->scheme),
-                            "duration_ms",
-                            duration_ms,
-                            "in_size",
-                            std::make_pair(static_cast<const size_t*>(data.node->length.data()),
-                                           data.node->length.size()),
-                            "total_size_bytes",
-                            total_size_bytes,
-                            "exec_GB_s",
-                            exec_bw,
-                            "max_mem_GB_s",
-                            max_memory_bw,
-                            "bw_efficiency_pct",
-                            efficiency_pct,
-                            "kernel_index",
-                            i);
-            }
-
-#ifdef REF_DEBUG
-            refLibOp.VerifyResult(&data);
-#endif
+            rocfft_cout << i << " ";
         }
+        rocfft_cout << std::endl;
+        rocfft_cout << "\tbatch:   " << execPlan.execSeq[i]->batch << std::endl;
+        rocfft_cout << "\tidist:   " << execPlan.execSeq[i]->iDist << std::endl;
+        rocfft_cout << "\todist:   " << execPlan.execSeq[i]->oDist << std::endl;
+        rocfft_cout << "\tistride:";
+        for(const auto& i : execPlan.execSeq[i]->inStride)
+        {
+            rocfft_cout << " " << i;
+        }
+        rocfft_cout << std::endl;
+        rocfft_cout << "\tostride:";
+        for(const auto& i : execPlan.execSeq[i]->outStride)
+        {
+            rocfft_cout << " " << i;
+        }
+        rocfft_cout << std::endl;
+
+        RefLibOp refLibOp(&data);
+#endif
+
+        // execution kernel:
+        if(emit_profile_log)
+            if(hipEventRecord(start) != hipSuccess)
+                throw std::runtime_error("hipEventRecord failure");
+
+        // give callback parameters to kernel launcher
+        data.callbacks = execPlan.execSeq[i]->callbacks;
+
+        // choose which compiled kernel to run
+        RTCKernel* localCompiledKernel = data.get_callback_type() == CallbackType::NONE
+                                             ? data.node->compiledKernel.get().get()
+                                             : data.node->compiledKernelWithCallbacks.get().get();
+
+        if(localCompiledKernel)
+            localCompiledKernel->launch(data, data.node->deviceProp);
         else
+            throw std::runtime_error("rocFFT null ptr function call error");
+
+        if(emit_profile_log)
+            if(hipEventRecord(stop) != hipSuccess)
+                throw std::runtime_error("hipEventRecord failure");
+
+        // If we were on the null stream, measure elapsed time
+        // and emit profile logging.  If a stream was given, we
+        // can't wait for the transform to finish, so we can't
+        // emit any information.
+        if(emit_profile_log)
         {
-            rocfft_cout << "null ptr function call error\n";
+            if(hipEventSynchronize(stop) != hipSuccess)
+                throw std::runtime_error("hipEventSynchronize failure");
+            size_t in_size_bytes
+                = data_size_bytes(data.node->length, data.node->precision, data.node->inArrayType);
+            size_t out_size_bytes
+                = data_size_bytes(data.node->length, data.node->precision, data.node->outArrayType);
+            size_t total_size_bytes = (in_size_bytes + out_size_bytes) * data.node->batch;
+
+            float duration_ms = 0.0f;
+            if(hipEventElapsedTime(&duration_ms, start, stop) != hipSuccess)
+                throw std::runtime_error("hipEventElapsedTime failure");
+            auto exec_bw        = execution_bandwidth_GB_per_s(total_size_bytes, duration_ms);
+            auto efficiency_pct = 0.0;
+            if(max_memory_bw != 0.0)
+                efficiency_pct = 100.0 * exec_bw / max_memory_bw;
+            if(processing_tuning)
+                tuningPacket->bw_effs[i] = efficiency_pct;
+
+            log_profile(__func__,
+                        "scheme",
+                        PrintScheme(execPlan.execSeq[i]->scheme),
+                        "duration_ms",
+                        duration_ms,
+                        "in_size",
+                        std::make_pair(static_cast<const size_t*>(data.node->length.data()),
+                                       data.node->length.size()),
+                        "total_size_bytes",
+                        total_size_bytes,
+                        "exec_GB_s",
+                        exec_bw,
+                        "max_mem_GB_s",
+                        max_memory_bw,
+                        "bw_efficiency_pct",
+                        efficiency_pct,
+                        "kernel_index",
+                        i);
         }
 
+#ifdef REF_DEBUG
+        refLibOp.VerifyResult(&data);
+#endif
+
         if(emit_kernelio_log && data.node->scheme != CS_KERNEL_CHIRP)
         {
             hipError_t err = hipPeekAtLastError();
@@ -898,7 +885,7 @@ void TransformPowX(const ExecPlan&       execPlan,
             {
                 *kernelio_stream << "Error: " << hipGetErrorName(err) << ", "
                                  << hipGetErrorString(err) << std::endl;
-                exit(-1);
+                throw std::runtime_error("hipPeekAtLastError failed");
             }
             if(hipDeviceSynchronize() != hipSuccess)
                 throw std::runtime_error("hipDeviceSynchronize failure");
@@ -911,17 +898,24 @@ void TransformPowX(const ExecPlan&       execPlan,
     {
         // offsets have only been applied to pointers given to kernels,
         // so apply them here for printing too
-        void* out_buffer_offset[2] = {out_buffer[0], out_buffer[1]};
+        void* out_buffer_offset[2] = {out_buffer[0], nullptr};
         if(execPlan.rootPlan->oOffset)
         {
             out_buffer_offset[0] = ptr_offset(out_buffer_offset[0],
                                               execPlan.rootPlan->oOffset,
                                               execPlan.rootPlan->precision,
                                               execPlan.rootPlan->outArrayType);
-            out_buffer_offset[1] = ptr_offset(out_buffer_offset[1],
-                                              execPlan.rootPlan->oOffset,
-                                              execPlan.rootPlan->precision,
-                                              execPlan.rootPlan->outArrayType);
+        }
+
+        if(array_type_is_planar(execPlan.rootPlan->outArrayType))
+        {
+            out_buffer_offset[1] = out_buffer[1];
+
+            if(execPlan.rootPlan->oOffset)
+                out_buffer_offset[1] = ptr_offset(out_buffer_offset[1],
+                                                  execPlan.rootPlan->oOffset,
+                                                  execPlan.rootPlan->precision,
+                                                  execPlan.rootPlan->outArrayType);
         }
 
         std::vector<hostbuf> bufOutHost;
