@@ -222,8 +222,8 @@ void stockham_combo(ComputeScheme                     scheme,
 
 void build_stockham_function_pool(CompileQueue& queue)
 {
-    // build everything in the function pool
-    function_pool& fp = function_pool::get_function_pool();
+    // build everything in the 64k LDS function pool
+    function_pool fp(65536);
 
     // fused Bluestein kernels are always built at runtime
     auto fuseBlue = BluesteinFuseType::BFT_NONE;
@@ -728,7 +728,7 @@ void build_solution_kernels(CompileQueue& queue)
 
 int main(int argc, char** argv)
 {
-    if(argc < 5)
+    if(argc < 4)
     {
         puts("Usage: rocfft_aot_helper temp_cachefile.db output_cachefile.db "
              "path/to/rocfft_rtc_helper gfx000 gfx001 ...");
