@@ -104,17 +104,16 @@ std::set<std::vector<unsigned int>> power_set(std::vector<unsigned int>::const_i
     return ret;
 }
 
-std::set<unsigned int>
+std::set<unsigned int, std::greater<unsigned int>>
     supported_threads_per_transform(const std::vector<unsigned int>& factorization)
 {
-    std::set<unsigned int> tpts;
-    auto                   tpt_candidates = power_set(factorization.begin(), factorization.end());
+    std::set<unsigned int, std::greater<unsigned int>> tpts;
+    auto tpt_candidates = power_set(factorization.begin(), factorization.end());
     for(auto tpt : tpt_candidates)
     {
         if(tpt.empty())
             continue;
-        auto product = std::accumulate(tpt.begin(), tpt.end(), 1U, std::multiplies<unsigned int>());
-        tpts.insert(product);
+        tpts.insert(product(tpt.begin(), tpt.end()));
     }
     return tpts;
 }
@@ -147,6 +146,7 @@ std::string test_kernel_name(unsigned int                     length,
 }
 
 std::string test_kernel_src(const std::string&               kernel_name,
+                            hipDeviceProp_t                  device_prop,
                             unsigned int&                    transforms_per_block,
                             unsigned int                     length,
                             ComputeScheme                    compute_scheme,
@@ -165,6 +165,8 @@ std::string test_kernel_src(const std::string&               kernel_name,
     specs.threads_per_transform = tpt;
     specs.half_lds              = half_lds;
     specs.direct_to_from_reg    = direct_to_from_reg;
+    // aim for occupancy-2
+    specs.lds_byte_limit = device_prop.sharedMemPerBlock / 2;
 
     return stockham_rtc(specs,
                         specs,
@@ -229,7 +231,6 @@ float launch_kernel(RTCKernel&             kernel,
     kargs.append_ptr(data.stride_in.data());
     kargs.append_ptr(data.stride_out.data());
     kargs.append_size_t(data.batch);
-    kargs.append_size_t(0);
     kargs.append_ptr(nullptr);
     kargs.append_ptr(nullptr);
     kargs.append_unsigned_int(0);
@@ -436,8 +437,12 @@ int main(int argc, char** argv)
         std::vector<unsigned int> best_factorization;
         std::string               best_kernel_src;
 
+        size_t count = 0;
         for(auto factorization : factorizations)
         {
+            ++count;
+            std::cout << "factorization " << count << " of " << factorizations.size() << std::endl;
+
             auto tpts = supported_threads_per_transform(factorization);
 
             // go through all permutations of the factors
@@ -447,68 +452,76 @@ int main(int argc, char** argv)
                 {
                     for(auto tpt : tpts)
                     {
-                        if(tpt < wgs)
+                        // tpt must fit into wgs
+                        if(tpt > wgs)
+                            continue;
+
+                        // There's no point in testing a half-full
+                        // wgs since so many threads will be doing
+                        // nothing.  And at larger wgs you could just
+                        // drop to a smaller one anyway.
+                        if(tpt <= wgs / 2)
+                            continue;
+
+                        for(bool half_lds : {true, false})
                         {
-                            for(bool half_lds : {true, false})
+                            for(bool direct_to_from_reg : {true, false})
                             {
-                                for(bool direct_to_from_reg : {true, false})
+                                // half lds currently requires direct to/from reg
+                                if(half_lds && !direct_to_from_reg)
+                                    continue;
+                                auto kernel_name = test_kernel_name(
+                                    length, factorization, wgs, tpt, half_lds, direct_to_from_reg);
+                                unsigned int transforms_per_block = 0;
+                                auto         kernel_src           = test_kernel_src(kernel_name,
+                                                                  device_prop,
+                                                                  transforms_per_block,
+                                                                  length,
+                                                                  compute_scheme,
+                                                                  precision,
+                                                                  factorization,
+                                                                  wgs,
+                                                                  tpt,
+                                                                  half_lds,
+                                                                  direct_to_from_reg);
+
+                                auto code = compile_inprocess(kernel_src, device_prop.gcnArchName);
+                                RTCKernelStockham kernel(kernel_name, code);
+
+                                float time = launch_kernel(
+                                    kernel,
+                                    DivRoundingUp<unsigned int>(data.batch, transforms_per_block),
+                                    tpt * transforms_per_block,
+                                    get_lds_bytes(length, transforms_per_block, half_lds),
+                                    ntrial,
+                                    device_prop,
+                                    data);
+
+                                // print median time for this length
+                                // in a format that can be easily
+                                // grepped for and shoved into a
+                                // database if desired
+                                std::cout << length << ", " << kernel_name << ", "
+                                          << std::setprecision(3) << static_cast<double>(time)
+                                          << "ms " << std::endl;
+
+                                if(time < best_time)
                                 {
-                                    // half lds currently requires direct to/from reg
-                                    if(half_lds && !direct_to_from_reg)
-                                        continue;
-                                    auto         kernel_name          = test_kernel_name(length,
-                                                                        factorization,
-                                                                        wgs,
-                                                                        tpt,
-                                                                        half_lds,
-                                                                        direct_to_from_reg);
-                                    unsigned int transforms_per_block = 0;
-                                    auto         kernel_src           = test_kernel_src(kernel_name,
-                                                                      transforms_per_block,
-                                                                      length,
-                                                                      compute_scheme,
-                                                                      precision,
-                                                                      factorization,
-                                                                      wgs,
-                                                                      tpt,
-                                                                      half_lds,
-                                                                      direct_to_from_reg);
-
-                                    auto code
-                                        = compile_inprocess(kernel_src, device_prop.gcnArchName);
-                                    RTCKernelStockham kernel(kernel_name, code);
-
-                                    float time = launch_kernel(
-                                        kernel,
-                                        DivRoundingUp<unsigned int>(data.batch,
-                                                                    transforms_per_block),
-                                        tpt * transforms_per_block,
-                                        get_lds_bytes(length, transforms_per_block, half_lds),
-                                        ntrial,
-                                        device_prop,
-                                        data);
-
-                                    // print median time for this length
-                                    // in a format that can be easily
-                                    // grepped for and shoved into a
-                                    // database if desired
-                                    std::cout << length << ", " << kernel_name << ", "
-                                              << std::setprecision(3) << static_cast<double>(time)
-                                              << std::endl;
-
-                                    if(time < best_time)
-                                    {
-                                        best_time               = time;
-                                        best_wgs                = wgs;
-                                        best_tpt                = tpt;
-                                        best_half_lds           = half_lds;
-                                        best_direct_to_from_reg = direct_to_from_reg;
-                                        best_factorization      = factorization;
-                                        best_kernel_src         = std::move(kernel_src);
-                                    }
+                                    best_time               = time;
+                                    best_wgs                = wgs;
+                                    best_tpt                = tpt;
+                                    best_half_lds           = half_lds;
+                                    best_direct_to_from_reg = direct_to_from_reg;
+                                    best_factorization      = factorization;
+                                    best_kernel_src         = std::move(kernel_src);
                                 }
                             }
                         }
+                        // found the largest number of threads
+                        // that'll fit into wgs, don't bother
+                        // testing fewer as that's just worse
+                        // utilization of the workgroup
+                        break;
                     }
                 }
             } while(std::next_permutation(factorization.begin(), factorization.end()));
@@ -579,6 +592,7 @@ int main(int argc, char** argv)
             unsigned int transforms_per_block = 0;
 
             auto kernel_src = test_kernel_src(kernel_name,
+                                              device_prop,
                                               transforms_per_block,
                                               length,
                                               compute_scheme,
