@@ -644,7 +644,6 @@ static std::string partial_pass_sbcc_64_64_64_rtc_body(const std::string& kernel
         const size_t *__restrict__ lengths,
         const size_t *__restrict__ stride,
         const size_t nbatch,
-        const unsigned int lds_padding,
         void *__restrict__ load_cb_fn,
         void *__restrict__ load_cb_data,
         unsigned int load_cb_lds_bytes,
@@ -710,8 +709,8 @@ static std::string partial_pass_sbcc_64_64_64_rtc_body(const std::string& kernel
     batch = blockIdx.x / plength;
     // offset = offset + batch * stride[dim]; // don't add batch here
     transform = lds_linear ? tile_index * 8 + threadIdx.x / 8 : tile_index * 8 + threadIdx.x % 8;
-    stride_lds = lds_linear ? 64 + (ebtype == EmbeddedType::NONE ? 0 : lds_padding)
-                            : 8 + (ebtype == EmbeddedType::NONE ? 0 : lds_padding);
+    stride_lds = lds_linear ? 64 + (ebtype == EmbeddedType::NONE ? 0 : 1)
+                            : 8 + (ebtype == EmbeddedType::NONE ? 0 : 1);
     stride_lds *= 4;
 
     offset_lds = lds_linear ? stride_lds * (transform % 8) : threadIdx.x % 8;
@@ -1132,7 +1131,6 @@ RTCKernelArgs RTCKernelPartialPassSBCC64Cubed::get_launch_args(DeviceCallIn& dat
     kargs.append_ptr(kargs_lengths(data.node->devKernArg));
     kargs.append_ptr(kargs_stride_in(data.node->devKernArg));
     kargs.append_size_t(data.node->batch);
-    kargs.append_size_t(data.node->lds_padding);
 
     // callback params
     kargs.append_ptr(data.callbacks.load_cb_fn);
