@@ -30,9 +30,11 @@
 #include "logging.h"
 #include "plan.h"
 #include "rocfft/rocfft.h"
+#include "rocfft_exception.h"
 #include "transform.h"
 
 rocfft_status rocfft_execution_info_create(rocfft_execution_info* info)
+try
 {
     rocfft_execution_info einfo = new rocfft_execution_info_t;
     *info                       = einfo;
@@ -40,8 +42,13 @@ rocfft_status rocfft_execution_info_create(rocfft_execution_info* info)
 
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 rocfft_status rocfft_execution_info_destroy(rocfft_execution_info info)
+try
 {
     log_trace(__func__, "info", info);
     if(info != nullptr)
@@ -49,10 +56,15 @@ rocfft_status rocfft_execution_info_destroy(rocfft_execution_info info)
 
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 rocfft_status rocfft_execution_info_set_work_buffer(rocfft_execution_info info,
                                                     void*                 work_buffer,
                                                     const size_t          size_in_bytes)
+try
 {
     log_trace(__func__, "info", info, "work_buffer", work_buffer, "size_in_bytes", size_in_bytes);
     if(!work_buffer)
@@ -62,18 +74,28 @@ rocfft_status rocfft_execution_info_set_work_buffer(rocfft_execution_info info,
 
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 rocfft_status rocfft_execution_info_set_stream(rocfft_execution_info info, void* stream)
+try
 {
     log_trace(__func__, "info", info, "stream", stream);
     info->rocfft_stream = (hipStream_t)stream;
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 rocfft_status rocfft_execution_info_set_load_callback(rocfft_execution_info info,
                                                       void**                cb_functions,
                                                       void**                cb_data,
                                                       size_t                shared_mem_bytes)
+try
 {
     // currently, we're not allocating LDS for callbacks, so fail
     // if any was requested
@@ -85,11 +107,16 @@ rocfft_status rocfft_execution_info_set_load_callback(rocfft_execution_info info
     info->callbacks.load_cb_lds_bytes = shared_mem_bytes;
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 rocfft_status rocfft_execution_info_set_store_callback(rocfft_execution_info info,
                                                        void**                cb_functions,
                                                        void**                cb_data,
                                                        size_t                shared_mem_bytes)
+try
 {
     // currently, we're not allocating LDS for callbacks, so fail
     // if any was requested
@@ -101,6 +128,10 @@ rocfft_status rocfft_execution_info_set_store_callback(rocfft_execution_info inf
     info->callbacks.store_cb_lds_bytes = shared_mem_bytes;
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 std::vector<size_t> rocfft_plan_t::MultiPlanTopologicalSort() const
 {
@@ -188,69 +219,90 @@ void rocfft_plan_t::LogSortedPlan(const std::vector<size_t>& sortedIdx) const
         return;
     }
 
-    // multi-device plan, log that with dependency graph
-    if(!LOG_GRAPH_ENABLED())
-        return;
-
-    auto& os = *LogSingleton::GetInstance().GetGraphOS();
-    os << "digraph plan {\n";
-    os << "ranksep=2;\n";
-
-    // gather up all of the groups for the nodes so they can turn into clustered subgraphs
-    std::multimap<std::string, size_t> groups;
-
-    for(auto i = sortedIdx.begin(); i != sortedIdx.end(); ++i)
+    if(LOG_PLAN_ENABLED())
     {
-        auto idx = *i;
-
-        const auto& antecedents = multiPlanAntecedents[idx];
-        for(auto antecedentIdx : antecedents)
+        auto& os = *LogSingleton::GetInstance().GetPlanOS();
+        os << "multiPlans: " << multiPlan.size() << "\n";
+        for(size_t midx = 0; midx < multiPlan.size(); ++midx)
         {
-            os << antecedentIdx << " -> " << idx << ";\n";
+            if(multiPlan[midx]->ExecutesOnLocalRank())
+            {
+                os << "multiPlan: " << midx << "\n";
+                os << "\tantecedents:";
+                const auto& antecedents = multiPlanAntecedents[midx];
+                for(auto antecedentIdx : antecedents)
+                {
+                    os << " " << antecedentIdx;
+                }
+                os << "\n";
+                multiPlan[midx]->Print(os, 1);
+            }
         }
+    }
 
-        rocfft_ostream item;
-        if(!multiPlan[idx])
-            item << "(null)";
-        else
-            multiPlan[idx]->Print(item, 1);
-        item << std::endl;
-        std::string itemStr = item.str();
-        // escape \n and "
-        std::string itemStrEscaped;
-        for(auto c : itemStr)
+    if(LOG_GRAPH_ENABLED())
+    {
+        // multi-device plan, log that with dependency graph
+        auto& os = *LogSingleton::GetInstance().GetGraphOS();
+        os << "digraph plan {\n";
+        os << "ranksep=2;\n";
+
+        // gather up all of the groups for the nodes so they can turn into clustered subgraphs
+        std::multimap<std::string, size_t> groups;
+
+        for(auto i = sortedIdx.begin(); i != sortedIdx.end(); ++i)
         {
-            if(c == '\"')
+            auto idx = *i;
+
+            const auto& antecedents = multiPlanAntecedents[idx];
+            for(auto antecedentIdx : antecedents)
             {
-                itemStrEscaped.push_back('\\');
-                itemStrEscaped.push_back('\"');
-            }
-            else if(c == '\n')
-            {
-                itemStrEscaped.push_back('\\');
-                itemStrEscaped.push_back('n');
+                os << antecedentIdx << " -> " << idx << ";\n";
             }
+
+            rocfft_ostream item;
+            if(!multiPlan[idx])
+                item << "(null)";
             else
-                itemStrEscaped.push_back(c);
-        }
-        os << idx << " [label=\"" << idx << "\\n"
-           << multiPlan[idx]->description << "\" tooltip=\"" << itemStrEscaped << "\"];\n";
+                multiPlan[idx]->Print(item, 1);
+            item << std::endl;
+            std::string itemStr = item.str();
+            // escape \n and "
+            std::string itemStrEscaped;
+            for(auto c : itemStr)
+            {
+                if(c == '\"')
+                {
+                    itemStrEscaped.push_back('\\');
+                    itemStrEscaped.push_back('\"');
+                }
+                else if(c == '\n')
+                {
+                    itemStrEscaped.push_back('\\');
+                    itemStrEscaped.push_back('n');
+                }
+                else
+                    itemStrEscaped.push_back(c);
+            }
+            os << idx << " [label=\"" << idx << "\\n"
+               << multiPlan[idx]->description << "\" tooltip=\"" << itemStrEscaped << "\"];\n";
 
-        groups.insert(std::make_pair(multiPlan[idx]->group, idx));
-    }
-    auto giter = groups.begin();
-    while(giter != groups.end())
-    {
-        auto gend = groups.upper_bound(giter->first);
-        os << "subgraph cluster_" << giter->first << " {\n";
-        os << "\nlabel=\"" << giter->first << "\";\n";
-        for(; giter != gend; ++giter)
-        {
-            os << giter->second << ";";
+            groups.insert(std::make_pair(multiPlan[idx]->group, idx));
         }
-        os << "}\n";
+        auto giter = groups.begin();
+        while(giter != groups.end())
+        {
+            auto gend = groups.upper_bound(giter->first);
+            os << "subgraph cluster_" << giter->first << " {\n";
+            os << "\nlabel=\"" << giter->first << "\";\n";
+            for(; giter != gend; ++giter)
+            {
+                os << giter->second << ";";
+            }
+            os << "}\n";
+        }
+        os << "}\n" << std::endl;
     }
-    os << "}\n" << std::endl;
 }
 
 void rocfft_plan_t::Execute(void* in_buffer[], void* out_buffer[], rocfft_execution_info info)
@@ -287,7 +339,7 @@ void rocfft_plan_t::Execute(void* in_buffer[], void* out_buffer[], rocfft_execut
 
     for(auto i = sortedIdx.begin(); i != sortedIdx.end(); ++i)
     {
-        auto idx = *i;
+        const auto idx = *i;
 
         if(!multiPlan[idx])
             continue;
@@ -296,22 +348,28 @@ void rocfft_plan_t::Execute(void* in_buffer[], void* out_buffer[], rocfft_execut
 
         for(auto antecedentIdx : multiPlanAntecedents[idx])
         {
+            // The multiPlan item may not be created on this rank (in which case it's not a
+            // rank-local operation):
             if(!multiPlan[antecedentIdx])
                 continue;
 
-            // check if antecedent involved us
+            // Check if antecedent involved us:
             auto& antecedent = *multiPlan[antecedentIdx];
 
-            // the antecedent involved us somehow, wait for it
+            // The antecedent involved us somehow, wait for it:
             if(antecedent.ExecutesOnRank(local_comm_rank))
+            {
                 antecedent.Wait();
+            }
         }
 
-        // done waiting for all our antecedents, so this item can now proceed
+        // Done waiting for all our antecedents, so this item can now proceed.
 
-        // launch this item async
+        // Launch this item async:
         if(item.ExecutesOnRank(local_comm_rank))
+        {
             item.ExecuteAsync(this, in_buffer, out_buffer, info, idx);
+        }
     }
 
     // finished executing all items, wait for outstanding work to complete
@@ -324,7 +382,9 @@ void rocfft_plan_t::Execute(void* in_buffer[], void* out_buffer[], rocfft_execut
 
         auto& item = *multiPlan[idx];
         if(item.ExecutesOnRank(local_comm_rank))
+        {
             item.Wait();
+        }
     }
 }
 
@@ -332,6 +392,7 @@ rocfft_status rocfft_execute(const rocfft_plan     plan,
                              void*                 in_buffer[],
                              void*                 out_buffer[],
                              rocfft_execution_info info)
+try
 {
     log_trace(
         __func__, "plan", plan, "in_buffer", in_buffer, "out_buffer", out_buffer, "info", info);
@@ -351,12 +412,12 @@ rocfft_status rocfft_execute(const rocfft_plan     plan,
         }
         return rocfft_status_failure;
     }
-    catch(rocfft_status e)
-    {
-        return e;
-    }
     return rocfft_status_success;
 }
+catch(...)
+{
+    return rocfft_handle_exception();
+}
 
 void ExecPlan::ExecuteAsync(const rocfft_plan     plan,
                             void*                 in_buffer[],
@@ -371,17 +432,10 @@ void ExecPlan::ExecuteAsync(const rocfft_plan     plan,
     if(info)
         exec_info = *info;
 
-    // allocate stream for async operations if necessary
-    // for a single-device plan, we don't need to create
-    // additional streams or events
-    if(mgpuPlan)
+    // use the local stream if user didn't provide one
+    if(mgpuPlan && !exec_info.rocfft_stream)
     {
-        if(!exec_info.rocfft_stream)
-        {
-            this->stream.alloc();
-            exec_info.rocfft_stream = this->stream;
-        }
-        event.alloc();
+        exec_info.rocfft_stream = this->stream;
     }
 
     // TransformPowX below needs in_buffer, out_buffer to work with.
@@ -477,7 +531,7 @@ void ExecPlan::Wait()
 {
     // for a single-device plan, we don't need to synchronize
     // events
-    if(mgpuPlan)
+    if(mgpuPlan && event)
     {
         if(hipEventSynchronize(event) != hipSuccess)
             throw std::runtime_error("hipEventSynchronize failed");
