@@ -1,4 +1,4 @@
-// Copyright (C) 2020 - 2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (C) 2020 - 2025 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -25,14 +25,12 @@
 #include "logging.h"
 #include "plan.h"
 #include "repo.h"
+#include "rocfft_mpi.h"
 #include "twiddles.h"
 
+#include <limits>
 #include <sstream>
 
-#ifdef ROCFFT_MPI_ENABLE
-#include <mpi.h>
-#endif
-
 struct rocfft_mp_request_t
 {
 #ifdef ROCFFT_MPI_ENABLE
@@ -107,7 +105,44 @@ FMKey LeafNode::GetKernelKey() const
 void LeafNode::GetKernelFactors()
 {
     FMKey key     = GetKernelKey();
-    kernelFactors = function_pool::get_kernel(key).factors;
+    kernelFactors = pool.get_kernel(key).factors;
+
+    // Hard-coded kernel factors for len 64x64x64 partial-pass
+    // TODO: Remove this hard-coded logic once
+    // partial-pass is integrated into the stockham generators.
+    if(scheme == CS_KERNEL_STOCKHAM && applyPartialPass)
+        kernelFactors = {8, 8};
+    if(scheme == CS_KERNEL_STOCKHAM_BLOCK_CC && applyPartialPass)
+        kernelFactors = {8, 8};
+}
+
+void LeafNode::GetKernelPartialPassFactors()
+{
+    // Hard-coded kernel partial-pass factors for len 64x64x64.
+    // TODO: Remove this hard-coded logic once
+    // partial-pass is integrated into the Stockham generators.
+    if(scheme == CS_KERNEL_STOCKHAM && applyPartialPass)
+    {
+        kernelFactorsPP = {16};
+        std::stringstream msg;
+        msg << "work in the off-dimension:" << std::endl;
+        msg << "\t     radix: [";
+        for(const auto factor : kernelFactorsPP)
+            msg << " " << factor;
+        msg << " ] pass(es) + Hadamard product with twiddle factors. \n";
+        comments.push_back(msg.str());
+    }
+    if(scheme == CS_KERNEL_STOCKHAM_BLOCK_CC && applyPartialPass)
+    {
+        kernelFactorsPP = {4};
+        std::stringstream msg;
+        msg << "work in the off-dimension:" << std::endl;
+        msg << "\t     local data transposition + radix: [";
+        for(const auto factor : kernelFactorsPP)
+            msg << " " << factor;
+        msg << " ] pass(es). \n";
+        comments.push_back(msg.str());
+    }
 }
 
 bool LeafNode::KernelCheck(std::vector<FMKey>& kernel_keys)
@@ -154,7 +189,7 @@ bool LeafNode::KernelCheck(std::vector<FMKey>& kernel_keys)
             // get sbrc_trans_type from assignedKey (for sbrc)
             sbrcTranstype = assignedKey.sbrcTrans;
 
-            function_pool::add_new_kernel(assignedKey);
+            pool.add_new_kernel(assignedKey);
             specified_key = std::make_unique<FMKey>(assignedKey);
         }
     }
@@ -163,7 +198,7 @@ bool LeafNode::KernelCheck(std::vector<FMKey>& kernel_keys)
     // Note that the check is trivial if we are using "specified_key"
     // since we definitly have the kernel, but not trivial if it's the auto-gen key
     FMKey key = GetKernelKey();
-    if(!function_pool::has_function(key))
+    if(!pool.has_function(key))
     {
         if(LOG_TRACE_ENABLED())
             (*LogSingleton::GetInstance().GetTraceOS()) << PrintMissingKernelInfo(key);
@@ -171,11 +206,15 @@ bool LeafNode::KernelCheck(std::vector<FMKey>& kernel_keys)
         return false;
     }
 
-    dir2regMode = (function_pool::get_kernel(key).direct_to_from_reg)
+    dir2regMode = (pool.get_kernel(key).direct_to_from_reg)
                       ? DirectRegType::TRY_ENABLE_IF_SUPPORT
                       : DirectRegType::FORCE_OFF_OR_NOT_SUPPORT;
 
     GetKernelFactors();
+
+    if(applyPartialPass)
+        GetKernelPartialPassFactors();
+
     return true;
 }
 
@@ -237,10 +276,10 @@ bool LeafNode::CreateDeviceResources()
     return CreateLargeTwdTable();
 }
 
-void LeafNode::SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)
+void LeafNode::SetupGridParam(GridParam& gp)
 {
     // derived classes setup the gp (bwd, wgs, lds, padding), funPtr
-    SetupGPAndFnPtr_internal(fnPtr, gp);
+    SetupGridParam_internal(gp);
 
     auto key = GetKernelKey();
 
@@ -248,9 +287,9 @@ void LeafNode::SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)
     gp.lds_bytes = lds * complex_type_size(precision);
     if(scheme == CS_KERNEL_STOCKHAM && ebtype == EmbeddedType::NONE)
     {
-        if(function_pool::has_function(key))
+        if(pool.has_function(key))
         {
-            auto kernel = function_pool::get_kernel(key);
+            auto kernel = pool.get_kernel(key);
 
             // NB:
             // Special case on specific arch:
@@ -269,16 +308,24 @@ void LeafNode::SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)
                 gp.lds_bytes /= 2;
         }
     }
-    // SBCC support half-lds conditionally
-    if((scheme == CS_KERNEL_STOCKHAM_BLOCK_CC)
-       && (dir2regMode == DirectRegType::TRY_ENABLE_IF_SUPPORT) && (ebtype == EmbeddedType::NONE))
+    if(scheme == CS_KERNEL_STOCKHAM_BLOCK_CC)
     {
-        if(function_pool::has_function(key))
+        // SBCC support half-lds conditionally
+        if((dir2regMode == DirectRegType::TRY_ENABLE_IF_SUPPORT) && (ebtype == EmbeddedType::NONE)
+           && pool.has_function(key))
         {
-            auto kernel = function_pool::get_kernel(key);
+            auto kernel = pool.get_kernel(key);
             if(kernel.half_lds)
                 gp.lds_bytes /= 2;
         }
+
+        auto apply_large_twd = (largeTwdBase > 0 && ltwdSteps > 0);
+        if(apply_large_twd && largeTwdBase < 8)
+        {
+            // append twiddle table to dynamic lds
+            auto kernel = pool.get_kernel(key);
+            gp.lds_bytes += twiddles_large_size;
+        }
     }
     // NB:
     //   SBCR / SBRC are not able to use half-lds due to both of them can't satisfy dir-to/from-registers at them same time.
@@ -299,7 +346,7 @@ void LeafNode::SetupGridParamAndFuncPtr(DevFnCall& fnPtr, GridParam& gp)
  *****************************************************/
 
 // grid params are set up by RTC
-void TransposeNode::SetupGPAndFnPtr_internal(DevFnCall& fnPtr, GridParam& gp) {}
+void TransposeNode::SetupGridParam_internal(GridParam& gp) {}
 
 void TreeNode::SetTransposeOutputLength()
 {
@@ -479,12 +526,12 @@ void CommPointToPoint::ExecuteAsync(const rocfft_plan     plan,
                                     size_t                multiPlanIdx)
 {
     rocfft_scoped_device dev(srcLocation.device);
-    stream.alloc();
-    event.alloc();
 
-    auto local_comm_rank = plan->get_local_comm_rank();
+    if(LOG_PLAN_ENABLED())
+    {
+        log_plan("CommPointToPoint\n");
+    }
 
-    auto memSize       = numElems * element_size(precision, arrayType);
     auto srcWithOffset = ptr_offset(
         srcPtr.get(in_buffer, out_buffer, local_comm_rank), srcOffset, precision, arrayType);
     auto destWithOffset = ptr_offset(
@@ -492,7 +539,8 @@ void CommPointToPoint::ExecuteAsync(const rocfft_plan     plan,
 
     if(srcLocation.comm_rank == destLocation.comm_rank)
     {
-        auto hiprt = hipSuccess;
+        const auto memSize = numElems * element_size(precision, arrayType);
+        auto       hiprt   = hipSuccess;
         if(srcLocation.device == destLocation.device)
         {
             hiprt = hipMemcpyAsync(
@@ -525,8 +573,8 @@ void CommPointToPoint::ExecuteAsync(const rocfft_plan     plan,
         {
             MPI_Request request;
             const auto  mpiret = MPI_Isend(srcWithOffset,
-                                          memSize,
-                                          MPI_BYTE,
+                                          numElems,
+                                          rocfft_type_to_mpi_type(precision, arrayType),
                                           destLocation.comm_rank,
                                           multiPlanIdx,
                                           plan->desc.mpi_comm,
@@ -542,8 +590,8 @@ void CommPointToPoint::ExecuteAsync(const rocfft_plan     plan,
         {
             MPI_Request request;
             const auto  mpiret = MPI_Irecv(destWithOffset,
-                                          memSize,
-                                          MPI_BYTE,
+                                          numElems,
+                                          rocfft_type_to_mpi_type(precision, arrayType),
                                           srcLocation.comm_rank,
                                           multiPlanIdx,
                                           plan->desc.mpi_comm,
@@ -563,7 +611,7 @@ void CommPointToPoint::Wait()
 {
     WaitCommRequests();
 
-    if(hipEventSynchronize(event) != hipSuccess)
+    if(event && hipEventSynchronize(event) != hipSuccess)
         throw std::runtime_error("hipEventSynchronize failed");
 }
 
@@ -591,15 +639,15 @@ void CommScatter::ExecuteAsync(const rocfft_plan     plan,
                                size_t                multiPlanIdx)
 {
     rocfft_scoped_device dev(srcLocation.device);
-    stream.alloc();
-    event.alloc();
 
-    auto local_comm_rank = plan->get_local_comm_rank();
+    if(LOG_PLAN_ENABLED())
+    {
+        log_plan("CommScatter\n");
+    }
 
     for(unsigned int opIdx = 0; opIdx < ops.size(); ++opIdx)
     {
-        const auto& op      = ops[opIdx];
-        auto        memSize = op.numElems * element_size(precision, arrayType);
+        const auto& op = ops[opIdx];
 
         auto srcWithOffset = ptr_offset(
             srcPtr.get(in_buffer, out_buffer, local_comm_rank), op.srcOffset, precision, arrayType);
@@ -611,6 +659,7 @@ void CommScatter::ExecuteAsync(const rocfft_plan     plan,
         hipError_t err = hipSuccess;
         if(op.destLocation.comm_rank == srcLocation.comm_rank)
         {
+            const auto memSize = op.numElems * element_size(precision, arrayType);
             if(local_comm_rank == op.destLocation.comm_rank)
             {
                 if(srcLocation.device == op.destLocation.device)
@@ -638,8 +687,8 @@ void CommScatter::ExecuteAsync(const rocfft_plan     plan,
             {
                 MPI_Request request;
                 const auto  mpiret = MPI_Isend(srcWithOffset,
-                                              memSize,
-                                              MPI_BYTE,
+                                              op.numElems,
+                                              rocfft_type_to_mpi_type(precision, arrayType),
                                               op.destLocation.comm_rank,
                                               GetOperationCommTag(multiPlanIdx, opIdx),
                                               plan->desc.mpi_comm,
@@ -655,8 +704,8 @@ void CommScatter::ExecuteAsync(const rocfft_plan     plan,
             {
                 MPI_Request request;
                 const auto  mpiret = MPI_Irecv(destWithOffset,
-                                              memSize,
-                                              MPI_BYTE,
+                                              op.numElems,
+                                              rocfft_type_to_mpi_type(precision, arrayType),
                                               srcLocation.comm_rank,
                                               GetOperationCommTag(multiPlanIdx, opIdx),
                                               plan->desc.mpi_comm,
@@ -673,8 +722,9 @@ void CommScatter::ExecuteAsync(const rocfft_plan     plan,
 #endif
         }
     }
-    // All work is enqueued to the stream, record the event on the stream
-    if(hipEventRecord(event, stream) != hipSuccess)
+    // All rank-local work is enqueued to the stream, record the
+    // event on the stream
+    if(event && hipEventRecord(event, stream) != hipSuccess)
         throw std::runtime_error("hipEventRecord failed");
 }
 
@@ -682,7 +732,7 @@ void CommScatter::Wait()
 {
     WaitCommRequests();
 
-    if(hipEventSynchronize(event) != hipSuccess)
+    if(event && hipEventSynchronize(event) != hipSuccess)
         throw std::runtime_error("hipEventSynchronize failed");
 }
 
@@ -716,10 +766,10 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
                               rocfft_execution_info info,
                               size_t                multiPlanIdx)
 {
-    streams.resize(ops.size());
-    events.resize(ops.size());
-
-    auto local_comm_rank = plan->get_local_comm_rank();
+    if(LOG_PLAN_ENABLED())
+    {
+        log_plan("CommGather\n");
+    }
 
     for(unsigned int opIdx = 0; opIdx < ops.size(); ++opIdx)
     {
@@ -728,10 +778,6 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
         auto&       event  = events[opIdx];
 
         rocfft_scoped_device dev(op.srcLocation.device);
-        stream.alloc();
-        event.alloc();
-
-        auto memSize = op.numElems * element_size(precision, arrayType);
 
         auto srcWithOffset  = ptr_offset(op.srcPtr.get(in_buffer, out_buffer, local_comm_rank),
                                         op.srcOffset,
@@ -745,6 +791,8 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
         hipError_t err = hipSuccess;
         if(destLocation.comm_rank == op.srcLocation.comm_rank)
         {
+            const auto memSize = op.numElems * element_size(precision, arrayType);
+
             if(local_comm_rank == destLocation.comm_rank)
             {
                 if(op.srcLocation.device == destLocation.device)
@@ -776,8 +824,8 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
             {
                 MPI_Request request;
                 auto        rcmpi = MPI_Isend(srcWithOffset,
-                                       memSize,
-                                       MPI_BYTE,
+                                       op.numElems,
+                                       rocfft_type_to_mpi_type(precision, arrayType),
                                        destLocation.comm_rank,
                                        GetOperationCommTag(multiPlanIdx, opIdx),
                                        plan->desc.mpi_comm,
@@ -790,8 +838,8 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
             {
                 MPI_Request request;
                 auto        rcmpi = MPI_Irecv(destWithOffset,
-                                       memSize,
-                                       MPI_BYTE,
+                                       op.numElems,
+                                       rocfft_type_to_mpi_type(precision, arrayType),
                                        op.srcLocation.comm_rank,
                                        GetOperationCommTag(multiPlanIdx, opIdx),
                                        plan->desc.mpi_comm,
@@ -803,9 +851,10 @@ void CommGather::ExecuteAsync(const rocfft_plan     plan,
 #endif
         }
 
-        // FIXME: we don't need events for MPI communications.
-        // All work for this stream is enqueued, record the event on the stream
-        if(hipEventRecord(event, stream) != hipSuccess)
+        // All work for this stream is enqueued, record the event on
+        // the stream.  The event is only allocated if necessary, so
+        // check it before recording.
+        if(event && hipEventRecord(event, stream) != hipSuccess)
             throw std::runtime_error("hipEventRecord failed");
     }
 }
@@ -816,7 +865,7 @@ void CommGather::Wait()
 
     for(const auto& event : events)
     {
-        if(hipEventSynchronize(event) != hipSuccess)
+        if(event && hipEventSynchronize(event) != hipSuccess)
             throw std::runtime_error("hipEventSynchronize failed");
     }
 }
@@ -845,14 +894,178 @@ void CommGather::Print(rocfft_ostream& os, const int indent) const
     }
 }
 
+void CommAllToAll::ExecuteAsync(const rocfft_plan     plan,
+                                void*                 in_buffer[],
+                                void*                 out_buffer[],
+                                rocfft_execution_info info,
+                                size_t                multiPlanIdx)
+{
+    // check that we have as many elems in our count/offset buffers as
+    // we have ranks
+    const size_t num_ranks = plan->get_local_comm_size();
+    if(sendOffsets.size() != num_ranks || sendCounts.size() != num_ranks
+       || recvOffsets.size() != num_ranks || recvCounts.size() != num_ranks)
+        throw std::runtime_error(
+            "CommAllToAll: number of counts/offsets does not match number of ranks");
+
+    if(LOG_PLAN_ENABLED())
+    {
+        log_plan("CommAllToAll: deciding between MPI_Ialltoall and MPI_Ialltoallv\n");
+    }
+
+#ifdef ROCFFT_MPI_ENABLE
+
+    const auto elem_size = element_size(precision, arrayType);
+
+    // check if uniform counts
+    auto count_matches_first = [&](size_t count) { return count == sendCounts[0]; };
+    bool uniform_counts = std::all_of(sendCounts.begin(), sendCounts.end(), count_matches_first)
+                          && std::all_of(recvCounts.begin(), recvCounts.end(), count_matches_first);
+
+    MPI_Request request;
+
+    if(uniform_counts)
+    {
+        if(LOG_PLAN_ENABLED())
+            log_plan("Using MPI_Ialltoall\n");
+
+        const int send_count_bytes = static_cast<int>(sendCounts[0] * elem_size);
+
+        const auto mpiret = MPI_Ialltoall(sendBuf.get(in_buffer, out_buffer, local_comm_rank),
+                                          send_count_bytes,
+                                          MPI_CHAR,
+                                          recvBuf.get(in_buffer, out_buffer, local_comm_rank),
+                                          send_count_bytes,
+                                          MPI_CHAR,
+                                          plan->desc.mpi_comm,
+                                          &request);
+
+        if(mpiret != MPI_SUCCESS)
+        {
+            char errmsg[MPI_MAX_ERROR_STRING];
+            int  errlen = 0;
+            MPI_Error_string(mpiret, errmsg, &errlen);
+
+            comm_status   = COMM_MPI_ERROR;
+            error_message = "MPI_Ialltoall failed on rank " + std::to_string(local_comm_rank) + ": "
+                            + std::string(errmsg);
+
+            return;
+        }
+    }
+    else
+    {
+        if(LOG_PLAN_ENABLED())
+            log_plan("Using MPI_Ialltoallv\n");
+
+        const int local_comm_rank = plan->get_local_comm_rank();
+
+        // MPI takes ints for everything, convert our size_t elements to int bytes
+        auto convertToInt = [](const std::vector<size_t>& src, std::vector<int>& dest) {
+            dest.reserve(src.size());
+            std::copy(src.begin(), src.end(), std::back_inserter(dest));
+        };
+
+        std::vector<int> intSendOffsets;
+        std::vector<int> intSendCounts;
+        std::vector<int> intRecvOffsets;
+        std::vector<int> intRecvCounts;
+        convertToInt(sendOffsets, intSendOffsets);
+        convertToInt(sendCounts, intSendCounts);
+        convertToInt(recvOffsets, intRecvOffsets);
+        convertToInt(recvCounts, intRecvCounts);
+
+        const auto mpiret = MPI_Ialltoallv(sendBuf.get(in_buffer, out_buffer, local_comm_rank),
+                                           intSendCounts.data(),
+                                           intSendOffsets.data(),
+                                           rocfft_type_to_mpi_type(precision, arrayType),
+                                           recvBuf.get(in_buffer, out_buffer, local_comm_rank),
+                                           intRecvCounts.data(),
+                                           intRecvOffsets.data(),
+                                           rocfft_type_to_mpi_type(precision, arrayType),
+                                           plan->desc.mpi_comm,
+                                           &request);
+
+        if(mpiret != MPI_SUCCESS)
+        {
+            char errmsg[MPI_MAX_ERROR_STRING];
+            int  errlen = 0;
+            MPI_Error_string(mpiret, errmsg, &errlen);
+
+            comm_status   = COMM_MPI_ERROR;
+            error_message = "MPI_Ialltoallv failed on rank " + std::to_string(local_comm_rank)
+                            + ": " + std::string(errmsg);
+
+            return;
+        }
+    }
+
+    comm_requests.push_back(request);
+
+#else
+    throw std::runtime_error("CommAllToAll not implemented");
+#endif
+}
+
+void CommAllToAll::Wait()
+{
+    WaitCommRequests();
+
+    if(comm_status == COMM_MPI_ERROR)
+    {
+        // Now it's safe to throw or return error
+        throw std::runtime_error(error_message);
+    }
+}
+
+void CommAllToAll::Print(rocfft_ostream& os, const int indent) const
+{
+    std::string indentStr;
+    int         i = indent;
+    while(i--)
+        indentStr += "    ";
+
+    auto printVec = [&os](const char* prefix, const std::vector<size_t>& vec) {
+        os << prefix << ": ";
+        for(auto val : vec)
+            os << val << " ";
+        os << "\n";
+    };
+
+    // determine whether counts are uniform
+    auto count_matches_first = [&](size_t count) { return count == sendCounts[0]; };
+    bool uniform_counts = std::all_of(sendCounts.begin(), sendCounts.end(), count_matches_first)
+                          && std::all_of(recvCounts.begin(), recvCounts.end(), count_matches_first);
+
+    os << indentStr << "CommAllToAll " << precision_name(precision) << " "
+       << PrintArrayType(arrayType) << (uniform_counts ? " (MPI_Ialltoall)" : " (MPI_Ialltoallv)")
+       << ":\n";
+
+    if(uniform_counts)
+    {
+        // alltoall: just print the count once
+        os << indentStr << " count_per_rank: " << sendCounts[0] << "\n";
+    }
+    else
+    {
+        // alltoallv: print full arrays
+        printVec("sendOffsets", sendOffsets);
+        printVec("sendCounts", sendCounts);
+        printVec("recvOffsets", recvOffsets);
+        printVec("recvCounts", recvCounts);
+    }
+}
+
 void ExecPlan::Print(rocfft_ostream& os, const int indent) const
 {
     std::string indentStr;
     int         i = indent;
     while(i--)
         indentStr += "    ";
+    os << indentStr << "MPI rank: " << local_comm_rank << "\n";
     os << indentStr << "ExecPlan:" << std::endl;
     os << indentStr << "  deviceID: " << location.device << std::endl;
+    os << indentStr << "  local_comm_rank:" << local_comm_rank << "\n";
     os << indentStr << "  commRanks:" << location.comm_rank << std::endl;
     if(inputPtr)
         os << indentStr << "  inputPtr: " << inputPtr.str() << std::endl;
