@@ -352,40 +352,42 @@ inline void execute_gpu_fft(Tparams&              params,
     auto fft_status = params.execute(pibuffer.data(), pobuffer.data());
     if(fft_status != fft_status_success)
         throw std::runtime_error("rocFFT plan execution failure");
-    // work around potential problem of following hipMemcpy
-    // not properly waiting for rocFFT's kernels to finish
-    if(hipDeviceSynchronize() != hipSuccess)
-        throw std::runtime_error("hipDeviceSynchronize after execute failed");
 
     // if not comparing, then just executing the GPU FFT is all we
     // need to do
     if(!fftw_compare)
         return;
 
-    // finalize a multi-GPU transform
-    params.multi_gpu_finalize(obuffer, pobuffer);
-
     ASSERT_TRUE(!gpu_output.empty()) << "no output buffers";
-    for(unsigned int idx = 0; idx < gpu_output.size(); ++idx)
+
+    // if output is in multiple bricks, collect it into the
+    // host buffer where the results need to go
+    if(!params.ofields.empty())
+        params.multi_gpu_finalize(gpu_output, obuffer, pobuffer);
+    else
     {
-        ASSERT_TRUE(gpu_output[idx].data() != nullptr)
-            << "output buffer index " << idx << " is empty";
-        auto hip_status = hipMemcpy(gpu_output[idx].data(),
-                                    pobuffer.at(idx),
-                                    gpu_output[idx].size(),
-                                    hipMemcpyDeviceToHost);
-        if(hip_status != hipSuccess)
+        // otherwise, copy directly from the device
+        for(unsigned int idx = 0; idx < gpu_output.size(); ++idx)
         {
-            ++n_hip_failures;
-            std::stringstream ss;
-            ss << "hipMemcpy failure";
-            if(skip_runtime_fails)
+            ASSERT_TRUE(gpu_output[idx].data() != nullptr)
+                << "output buffer index " << idx << " is empty";
+            auto hip_status = hipMemcpy(gpu_output[idx].data(),
+                                        pobuffer.at(idx),
+                                        gpu_output[idx].size(),
+                                        hipMemcpyDeviceToHost);
+            if(hip_status != hipSuccess)
             {
-                throw ROCFFT_SKIP{ss.str()};
-            }
-            else
-            {
-                throw ROCFFT_FAIL{ss.str()};
+                ++n_hip_failures;
+                std::stringstream ss;
+                ss << "hipMemcpy failure";
+                if(skip_runtime_fails)
+                {
+                    throw ROCFFT_SKIP{ss.str()};
+                }
+                else
+                {
+                    throw ROCFFT_FAIL{ss.str()};
+                }
             }
         }
     }
@@ -1262,6 +1264,9 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
         pobuffer[i] = obuffer->at(i).data();
     }
 
+    // scatter data out to multi-GPUs if this is a multi-GPU test
+    params.multi_gpu_prepare(cpu_input, ibuffer, pibuffer, pobuffer);
+
     // Run CPU transform
     //
     // NOTE: This must happen after input is copied to GPU and input
@@ -1300,9 +1305,6 @@ inline void fft_vs_reference_impl(Tparams& params, bool round_trip)
             }
         });
 
-    // scatter data out to multi-GPUs if this is a multi-GPU test
-    params.multi_gpu_prepare(ibuffer, pibuffer, pobuffer);
-
     // execute GPU transform
     std::vector<hostbuf> gpu_output
         = allocate_host_buffer(params.precision, params.otype, params.osize);
